from .utilities.util import *
from .model.model import Cell2Clusters, Cell2Spots, Cell2VisiumSpots, Cell2SpotsStrips,Cell2SpotsEnhance
from .model.ot_model import solve_OT
from .Polyomino_argument import argument_parser
import pandas as pd
from scipy.sparse import issparse
import numpy as np
import scanpy as sc
import tqdm
import torch
from sklearn.metrics.pairwise import cosine_similarity as cos_s

class Polyomino:
    """
    This class runs the Polyomino algorithm on scRNA-seq data and spatial transcriptomics or bulk RNA datasets.
    It reconstructs spatial transcriptomics data based on the Polyomino algorithm.

    Attributes:
        scdata (anndata.AnnData): scRNA-seq data.
        stdata (anndata.AnnData): Spatial transcriptomics data.
        bulkX (anndata.AnnData): Bulk transcriptomics data (X-axis).
        bulkY (anndata.AnnData): Bulk transcriptomics data (Y-axis).
        genes (list): List of genes to use for mapping.
        histology (np.array): Histology data.
        cluster_time (int): Number of clustering iterations.
        custom_label (str): Custom label for clustering.
        pca (str): PCA method to use.
        n_pcs (int): Number of principal components.
        pc_frac (float): Fraction of PCs to use.
        target_num (int): Number of clusters.
        cluster_thres (float): Threshold to filter small values in clustering.
        ot_cluster_thres (float): Threshold for OT clustering.
        emptygrid (np.array): Empty grid matrix.
        shape (str): Shape of spatial transcriptomics data ('square' or 'hexagon').
        device (str): Device for training ('cpu' or 'cuda').
        cluster_matrix (np.ndarray): Matrix for clustering results.
        spot_matrix (np.ndarray): Matrix for spot mapping results.
    """

    def __init__(
        self, 
        scdata, 
        stdata=None,
        bulkX=None, 
        bulkY=None, 
        genes=None, 
        histology=None, 
        cluster_time=0, 
        custom_label=None,
        pca=None, 
        n_pcs=50, 
        pc_frac=1.0, 
        target_num=10,
        cluster_thres=None,
        ot_cluster_thres=0.01,
        emptygrid=None,
        shape="square", 
        device='cpu'
    ):
        """
        Initializes the Polyomino class with provided parameters.

        Args:
            scdata (anndata.AnnData): scRNA-seq data.
            stdata (anndata.AnnData): Spatial transcriptomics data.
            bulkX (anndata.AnnData): Bulk transcriptomics data (X-axis).
            bulkY (anndata.AnnData): Bulk transcriptomics data (Y-axis).
            genes (list): List of genes to use for mapping.
            histology (np.array): Histology data.
            cluster_time (int): Number of clustering iterations.
            custom_label (str): Custom label for clustering.
            pca (str): PCA method to use.
            n_pcs (int): Number of principal components.
            pc_frac (float): Fraction of PCs to use.
            target_num (int): Number of clusters.
            cluster_thres (float): Threshold to filter small values in clustering.
            ot_cluster_thres (float): Threshold for OT clustering.
            emptygrid (np.array): Empty grid matrix.
            shape (str): Shape of spatial transcriptomics data ('square' or 'hexagon').
            device (str): Device for training ('cpu' or 'cuda').
        """
        self.scdata = scdata
        self.stdata = stdata
        self.bulkX = bulkX
        self.bulkY = bulkY
        self.genes = genes
        self.histology = histology
        self.cluster_time = cluster_time
        self.custom_label = custom_label
        self.pca = pca
        self.n_pcs = n_pcs
        self.pc_frac = pc_frac
        self.target_num = target_num
        self.cluster_thres = cluster_thres
        self.ot_cluster_thres = ot_cluster_thres
        self.emptygrid = emptygrid
        self.shape = shape
        self.device = device

        # Initialize matrices
        self.cluster_matrix = None
        self.spot_matrix = None

    def allocate(self,num_epochs=500):
        """
        Runs the Polyomino algorithm for mapping scRNA-seq data to spatial transcriptomics or bulk RNA datasets.
        Generates a matrix with probabilities of each cell belonging to specific clusters.
        """
        if self.stdata is not None:
            self.bulkX = generate_Xstrips(self.stdata)
            self.bulkY = generate_Ystrips(self.stdata)
            self.cluster_label = []

            if self.cluster_time > 0 and self.custom_label:
                self.cluster_label.append(self.custom_label)
                self.cluster_matrix = cluster_mapping(
                    self.scdata, self.stdata, self.genes, label=self.custom_label, device=self.device, thres=self.cluster_thres, num_epochs=num_epochs
                )
                torch.cuda.empty_cache()

                if self.cluster_thres is not None:
                    self.cluster_matrix[self.cluster_matrix < self.cluster_thres] = -1e15
            else:
                if self.cluster_time == 0:
                    print("Starting spot mapping.")
                    self.spot_matrix = spot_mapping(self.scdata, self.stdata, genes=self.genes, device=self.device, num_epochs=num_epochs)
                else:
                    self.cluster_matrix = 0

            if self.cluster_time > 0:
                for i in range(self.cluster_time):
                    label = f'clutimes_{i+1}'
                    self.cluster_label.append(label)
                    print(f"Running {i+1}-th clustering.")
                    random_cluster(
                        self.stdata, label, embed=self.pca, n_pcs=self.n_pcs, pc_frac=self.pc_frac,
                        samples_time=self.cluster_time, shape=self.shape, target_num=self.target_num
                    )
                    print(f"Running {i+1}-th mapping.")
                    self.cluster_matrix += cluster_mapping(
                        self.scdata, self.stdata, self.genes, label=label, device=self.device, thres=self.cluster_thres, num_epochs=num_epochs
                    )
                    torch.cuda.empty_cache()
                self.cluster_matrix /= self.cluster_time

                if self.cluster_thres is not None:
                    self.cluster_matrix[self.cluster_matrix < self.cluster_thres] = -1e15

                print("Starting spot mapping.")
                self.spot_matrix = spot_mapping(
                    self.scdata, self.stdata, self.cluster_matrix.values, genes=self.genes, device=self.device, num_epochs=num_epochs
                )
        else:
            print("Starting spot mapping.")
            self.spot_matrix = spot_mapping(
                self.scdata, None, self.cluster_matrix.values, self.bulkX, self.bulkY, genes=self.genes, device=self.device, num_epochs=num_epochs
            )

        self.scdata.obs['grid_prob'] = np.max(self.spot_matrix, axis=1)
        print("Spot mapping completed.")
        print("Mapping matrix saved in zm.spot_matrix")

    def ot_allocate(self):
        """
        Runs the optimal transport algorithm for mapping scRNA-seq data to spatial transcriptomics or bulk RNA datasets.
        """
        if self.stdata is not None:
            self.bulkX = sc.AnnData(X=generate_Xstrips(self.stdata), var=self.stdata.var)
            self.bulkY = sc.AnnData(X=generate_Ystrips(self.stdata), var=self.stdata.var)
            self.scLocX = strips_mapping(self.scdata, self.bulkX, self.genes, device=self.device, thres=self.ot_cluster_thres)
            self.scLocY = strips_mapping(self.scdata, self.bulkY, self.genes, device=self.device, thres=self.ot_cluster_thres)
            self.spot_matrix = solve_OT(self.emptygrid, self.scLocX, self.scLocY, thres=self.ot_cluster_thres, njob=8)
        else:
            self.scLocX = strips_mapping(self.scdata, self.bulkX, self.genes, device=self.device, thres=self.ot_cluster_thres)
            self.scLocY = strips_mapping(self.scdata, self.bulkY, self.genes, device=self.device, thres=self.ot_cluster_thres)
            self.spot_matrix = solve_OT(self.emptygrid, self.scLocX, self.scLocY, thres=self.ot_cluster_thres, njob=8)


class Polyomino3D:
    """
    This class runs the Polyomino algorithm on scRNA-seq data and Visium spatial transcriptomics datasets.
    It reconstructs new spatial transcriptomics data using the Polyomino algorithm.
    Attributes:
        scdata (anndata.AnnData): scRNA-seq data.
        stdata (anndata.AnnData): 3D spatial transcriptomics data.
        genes (list): List of genes to use for mapping.
        cluster_time (int): Number of clustering iterations.
        custom_label (str): Custom label for clustering.
        pca (str): PCA method to use.
        n_pcs (int): Number of principal components.
        pc_frac (float): Fraction of PCs to use.
        target_num (int): Number of clusters.
        cluster_thres (float): Threshold to filter small values in clustering.
        shape (str): Shape of the spatial transcriptomics data ('square' or 'hexagon').
        device (str): Device for training ('cpu' or 'cuda').
        cluster_matrix (np.ndarray): Matrix for clustering results.
        spot_matrix (np.ndarray): Matrix for spot mapping results.
    """

    def __init__(
        self, 
        scdata, 
        stdata,
        genes=None, 
        cluster_time=1, 
        custom_label=None,
        pca=None, 
        n_pcs=50, 
        pc_frac=0.5, 
        target_num=10,
        cluster_thres=None,
        shape="hexagon", 
        device='cpu'
    ):
        """
        Initializes the Polyomino3D class with provided parameters.

        Args:
            scdata (anndata.AnnData): scRNA-seq data.
            stdata (anndata.AnnData): 3D spatial transcriptomics data.
            genes (list): List of genes to use for mapping.
            cluster_time (int): Number of clustering iterations.
            custom_label (str): Custom label for clustering.
            pca (str): PCA method to use.
            n_pcs (int): Number of principal components.
            pc_frac (float): Fraction of PCs to use.
            target_num (int): Number of clusters.
            cluster_thres (float): Threshold to filter small values in clustering.
            shape (str): Shape of the spatial transcriptomics data ('square' or 'hexagon').
            device (str): Device for training ('cpu' or 'cuda').
        """
        self.scdata = scdata
        self.stdata = stdata
        self.genes = genes
        self.cluster_time = cluster_time
        self.custom_label = custom_label
        self.pca = pca
        self.n_pcs = n_pcs
        self.pc_frac = pc_frac
        self.target_num = target_num
        self.cluster_thres = cluster_thres
        self.shape = shape
        self.device = device

        # Initialize matrices
        self.cluster_matrix = None
        self.spot_matrix = None

    def allocate(self, num_epochs=500):
        """
        Runs the Polyomino algorithm for mapping scRNA-seq data to 3D spatial transcriptomics data.
        Generates a matrix with probabilities of each cell belonging to specific clusters.
        """
        self.bulkX = generate_Xstrips(self.stdata)
        self.bulkY = generate_Ystrips(self.stdata)
        self.bulkZ = generate_Zstrips(self.stdata)
        self.cluster_label = []

        if self.cluster_time > 0 and self.custom_label is not None:
            self.cluster_label.append(self.custom_label)
            self.cluster_matrix = cluster_mapping(
                self.scdata, self.stdata, self.genes, label=self.custom_label, device=self.device, thres=self.cluster_thres, num_epochs=num_epochs
            )
            torch.cuda.empty_cache()

            if self.cluster_thres is not None:
                self.cluster_matrix[self.cluster_matrix < self.cluster_thres] = -1e15
        else:
            if self.cluster_time == 0:
                print("Starting spot mapping.")
                self.spot_matrix = spot_mapping3D(self.scdata, self.stdata, genes=self.genes, device=self.device, num_epochs=num_epochs)
            else:
                self.cluster_matrix = 0

        if self.cluster_time > 0:
            for i in range(self.cluster_time):
                label = f'clutimes_{i}'
                self.cluster_label.append(label)
                print(f"Running {i}-th clustering.")
                random_cluster(
                    self.stdata, label, embed=self.pca, n_pcs=self.n_pcs, pc_frac=self.pc_frac,
                    samples_time=self.cluster_time, shape=self.shape, target_num=self.target_num
                )
                print(f"Running {i}-th mapping.")
                self.cluster_matrix += cluster_mapping(
                    self.scdata, self.stdata, self.genes, label=label, device=self.device, thres=self.cluster_thres, num_epochs=num_epochs
                )
                torch.cuda.empty_cache()

            self.cluster_matrix /= self.cluster_time

            if self.cluster_thres is not None:
                self.cluster_matrix[self.cluster_matrix < self.cluster_thres] = -1e15

            print("Starting spot mapping.")
            self.spot_matrix = spot_mapping3D(
                self.scdata, self.stdata, self.cluster_matrix.values, genes=self.genes, device=self.device, num_epochs=num_epochs
            )
        else:
            self.spot_matrix = spot_mapping3D(
                self.scdata, self.stdata, self.cluster_matrix.values, genes=self.genes, device=self.device, num_epochs=num_epochs
            )

        self.scdata.obs['grid_prob'] = np.max(self.spot_matrix, axis=1)
        print("Spot mapping completed.")
        print("Mapping matrix saved in zm.spot_matrix")


class Polyomino_custom:
    """
    This class runs a customized Polyomino algorithm on scRNA-seq data and spatial transcriptomics data.
    It uses a model-based approach to reconstruct single cell resolution spatial transcriptomics data.

    Attributes:
        scdata (anndata.AnnData): scRNA-seq data.
        stdata (anndata.AnnData): Spatial transcriptomics data.
        bulkX (anndata.AnnData): Bulk transcriptomics data (X strips).
        bulkY (anndata.AnnData): Bulk transcriptomics data (Y strips).
        genes (list): Genes to use for mapping.
        custom_label (str): Custom label for clustering.
        learning_rate (float): Learning rate for the optimization process.
        num_epochs (int): Number of epochs for training.
        lambda_gx1 (float): Regularization parameter for Gx1.
        lambda_gy1 (float): Regularization parameter for Gy1.
        lambda_r (float): Regularization parameter for the residual term (optional).
        device (str): Device for training ('cpu' or 'cuda').
        spot_matrix (np.ndarray): Matrix of spot mappings.
    """

    def __init__(
        self, 
        scdata, 
        stdata=None,
        genes=None, 
        custom_label=None,
        learning_rate=0.01,
        num_epochs=500,
        lambda_gx1=1,
        lambda_gy1=1,
        lambda_r=None,
        device='cpu'
    ):
        """
        Initializes the Polyomino_custom class with provided parameters.

        Args:
            scdata (anndata.AnnData): scRNA-seq data.
            stdata (anndata.AnnData): Spatial transcriptomics data.
            genes (list): List of genes to use for mapping.
            custom_label (str): Custom label for clustering.
            learning_rate (float): Learning rate for optimization.
            num_epochs (int): Number of epochs for training.
            lambda_gx1 (float): Regularization parameter for Gx1.
            lambda_gy1 (float): Regularization parameter for Gy1.
            lambda_r (float): Regularization parameter for the residual term (optional).
            device (str): Device for training ('cpu' or 'cuda').
        """
        self.scdata = scdata
        self.stdata = stdata
        self.bulkX = None
        self.bulkY = None
        self.genes = genes
        self.custom_label = custom_label
        self.spot_matrix = None
        self.learning_rate = learning_rate
        self.num_epochs = num_epochs
        self.lambda_gx1 = lambda_gx1
        self.lambda_gy1 = lambda_gy1
        self.lambda_r = lambda_r
        self.device = device

    def allocate(self):
        """
        Runs the customized Polyomino algorithm on the scRNA-seq and spatial transcriptomics datasets.
        Produces a matrix with probabilities of each cell belonging to grids.
        """
        # Generate bulk data strips
        self.bulkX = sc.AnnData(X=generate_Xstrips(self.stdata), var=self.stdata.var)
        self.bulkY = sc.AnnData(X=generate_Ystrips(self.stdata), var=self.stdata.var)

        # Preprocess scRNA-seq data and get overlap genes
        preprocess(self.scdata, self.bulkX, self.genes)
        overlap_genes = self.scdata.uns["overlap"]

        # Extract data matrices
        S = np.array(self.scdata[:, overlap_genes].X.toarray(), dtype="float32")
        Gx1 = np.array(self.bulkX[:, overlap_genes].X, dtype="float32")
        Gx2 = np.array(self.bulkY[:, overlap_genes].X, dtype="float32")
        ST = self.stdata[:, overlap_genes]

        # Run enhanced cell-to-spot mapping
        self.spot_matrix = Cell2SpotsEnhance(
            S=S, ST=ST, Gx=Gx1, Gy=Gx2,
            lambda_gx1=self.lambda_gx1, lambda_gy1=self.lambda_gy1,
            custom_regions=self.custom_label, lambda_r=self.lambda_r,
            device=self.device
        ).fit(
            learning_rate=self.learning_rate, num_epochs=self.num_epochs, print_each=100
        )

        # Update scdata with the maximum spot probability
        self.scdata.obs['grid_prob'] = np.max(self.spot_matrix, axis=1)
        print("Finish spot mapping.")

def strips_mapping(
    scadata,
    bulk,
    genes=None,
    device='cpu',
    num_epochs=500,
    learning_rate=0.1,
    thres=None
):
    """
    Maps scRNA-seq data to bulk transcriptomics data using a model-based approach.

    Args:
        scadata (anndata.AnnData): scRNA-seq data.
        bulk (anndata.AnnData): Bulk transcriptomics data.
        genes (list, optional): List of genes to use for mapping. If None, all genes are used.
        device (str, optional): Device for training ('cpu' or 'cuda'). Default is 'cpu'.
        num_epochs (int, optional): Number of epochs for training. Default is 500.
        learning_rate (float, optional): Learning rate for training. Default is 0.1.
        thres (float, optional): Threshold to filter small values in the output matrix. If None, no filtering is applied.

    Returns:
        np.ndarray: Mapping matrix from scRNA-seq to bulk transcriptomics data.
    """
    # Preprocess data with the specified genes, if provided
    if genes is None:
        preprocess(scadata, bulk)
    else:
        preprocess(scadata, bulk, genes)
    
    # Get overlapping genes
    overlap_genes = scadata.uns["overlap"]
    
    # Convert scRNA-seq and bulk data to numpy arrays
    S = np.array(scadata[:, overlap_genes].X.toarray(), dtype="float32")
    G = np.array(bulk[:, overlap_genes].X, dtype="float32")
    
    # Initialize the cell-to-clusters mapper
    bulk_mapper = Cell2Clusters(scdata=S, clusters=G, device=device)
    
    # Fit the model to get the mapping matrix
    bulk_matrix = bulk_mapper.fit(num_epochs=num_epochs, learning_rate=learning_rate, print_each=100)
    
    # Apply threshold filtering if specified
    if thres is not None:
        bulk_matrix[bulk_matrix < thres] = 0
    return bulk_matrix

def cluster_mapping(
    scadata,
    stdata,
    genes=None,
    label='leiden',
    device='cpu',
    num_epochs=500,
    learning_rate=0.1,
    thres=None
):
    """
    Maps scRNA-seq data to specific regions based on spatial transcriptomics cluster labels.

    Args:
        scadata (anndata.AnnData): scRNA-seq data.
        stdata (anndata.AnnData): Spatial transcriptomics data.
        genes (list, optional): List of genes to use for mapping. If None, all genes are used.
        label (str): Column name in stdata.obs containing cluster labels.
        device (str, optional): Device for training ('cpu' or 'cuda'). Default is 'cpu'.
        num_epochs (int, optional): Number of epochs for training. Default is 500.
        learning_rate (float, optional): Learning rate for training. Default is 0.1.
        thres (float, optional): Threshold to filter small values in the output matrix. If None, no filtering is applied.

    Returns:
        np.ndarray: Mapping matrix from scRNA-seq to cluster transcriptomics data.
    """
    # Generate cluster expression data from spatial transcriptomics data
    cluster_bulk = generate_cluster_exp(stdata, label=label)
    cluster_bulk = sc.AnnData(X=cluster_bulk, var=stdata.var)
    
    # Preprocess the data with the specified genes, if provided
    if genes is None:
        preprocess(scadata, cluster_bulk)
    else:
        preprocess(scadata, cluster_bulk, genes)
    
    # Get overlapping genes
    overlap_genes = scadata.uns["overlap"]
    
    # Convert scRNA-seq and cluster data to numpy arrays
    scdata_arr = np.array(scadata[:, overlap_genes].X.toarray(), dtype="float32")
    clusters_arr = np.array(cluster_bulk[:, overlap_genes].X, dtype="float32")
    
    # Initialize the cell-to-clusters mapper
    cluster_mapper = Cell2Clusters(scdata=scdata_arr, clusters=clusters_arr, device=device)
    
    # Fit the model to get the mapping matrix
    cluster_mapper_matrix = cluster_mapper.fit(num_epochs=num_epochs, learning_rate=learning_rate, print_each=100)
    
    # Transform the cluster labels
    cell_clu = one_hot_encode_labels(stdata.obs[label])
    
    # Compute the mapping matrix
    cell_spot = cluster_mapper_matrix @ cell_clu.T
    
    # Apply threshold filtering if specified
    if thres is not None:
        return (cell_spot > thres).astype(int)
    
    return cell_spot

def spot_mapping(
    scadata,
    stdata,
    cluster_mapping_matrix=None,
    x_bulk=None,
    y_bulk=None,
    genes=None,
    device='cpu',
    num_epochs=500,
    learning_rate=0.1
):
    """
    Maps single-cell RNA-seq data to spatial spots using spatial transcriptomics data.

    Args:
        scadata (anndata.AnnData): scRNA-seq data.
        stdata (anndata.AnnData): Spatial transcriptomics data.
        cluster_mapping_matrix (np.ndarray, optional): Matrix mapping scRNA-seq data to clusters.
        x_bulk (anndata.AnnData, optional): Bulk transcriptomics data in X direction. If None, generated from stdata.
        y_bulk (anndata.AnnData, optional): Bulk transcriptomics data in Y direction. If None, generated from stdata.
        genes (list, optional): List of genes to use for mapping. If None, all overlapping genes are used.
        device (str, optional): Device for training ('cpu' or 'cuda'). Default is 'cpu'.
        num_epochs (int, optional): Number of epochs for training. Default is 500.
        learning_rate (float, optional): Learning rate for training. Default is 0.1.

    Returns:
        np.ndarray: Mapping matrix from single cells to spatial spots.
    """
    # Generate bulk data if not provided
    if x_bulk is None and y_bulk is None:
        x_bulk = sc.AnnData(X=generate_Xstrips(stdata), var=stdata.var)
        y_bulk = sc.AnnData(X=generate_Ystrips(stdata), var=stdata.var)
    
    # Preprocess the scRNA-seq data
    preprocess(scadata, x_bulk, genes)
    
    # Get overlapping genes
    overlap_genes = scadata.uns["overlap"]
    
    # Convert data to numpy arrays
    S = np.array(scadata[:, overlap_genes].X.toarray(), dtype="float32")
    Gx1 = np.array(x_bulk[:, overlap_genes].X, dtype="float32")
    Gx2 = np.array(y_bulk[:, overlap_genes].X, dtype="float32")
    
    # Perform mapping to spatial spots
    if stdata is None:
        # Use cell2spots_strips if spatial data is not provided
        mapping_matrix = Cell2SpotsStrips(
            S=S,
            Gx=Gx1,
            Gy=Gx2,
            cluster_matrix=cluster_mapping_matrix,
            device=device
        ).fit(learning_rate=learning_rate, num_epochs=num_epochs, print_each=100)
    else:
        # Use cell2spots if spatial data is provided
        ST = stdata[:, overlap_genes]
        mapping_matrix = Cell2Spots(
            S=S,
            ST=ST,
            Gx=Gx1,
            Gy=Gx2,
            cluster_matrix=cluster_mapping_matrix,
            device=device
        ).fit(learning_rate=learning_rate, num_epochs=num_epochs, print_each=100)
    
    return mapping_matrix

def spot_mapping3D(
    scadata,
    stdata,
    cluster_mapping_matrix=None,
    genes=None,
    device='cpu',
    num_epochs=500,
    learning_rate=0.1
):
    """
    Maps single-cell RNA-seq data to spatial spots in a 3D spatial transcriptomics setup.

    Args:
        scadata (anndata.AnnData): scRNA-seq data.
        stdata (anndata.AnnData): Spatial transcriptomics data.
        cluster_mapping_matrix (np.ndarray, optional): Matrix mapping scRNA-seq data to clusters.
        genes (list, optional): List of genes to use for mapping. If None, all overlapping genes are used.
        device (str, optional): Device for training ('cpu' or 'cuda'). Default is 'cpu'.
        num_epochs (int, optional): Number of epochs for training. Default is 500.
        learning_rate (float, optional): Learning rate for training. Default is 0.1.

    Returns:
        np.ndarray: Mapping matrix from single cells to spatial spots.
    """
    # Generate bulk data for X, Y, and Z directions
    x_bulk = generate_Xstrips(stdata)
    y_bulk = generate_Ystrips(stdata)
    z_bulk = generate_Zstrips(stdata)
    
    # Create AnnData objects for bulk data
    bx1 = sc.AnnData(X=x_bulk, var=stdata.var)
    bx2 = sc.AnnData(X=y_bulk, var=stdata.var)
    bx3 = sc.AnnData(X=z_bulk, var=stdata.var)
    
    # Preprocess scRNA-seq data
    preprocess(scadata, bx1, genes)
    
    # Get overlapping genes
    overlap_genes = scadata.uns["overlap"]
    
    # Convert data to numpy arrays
    S = np.array(scadata[:, overlap_genes].X.toarray(), dtype="float32")
    Gx1 = np.array(bx1[:, overlap_genes].X, dtype="float32")
    Gx2 = np.array(bx2[:, overlap_genes].X, dtype="float32")
    Gx3 = np.array(bx3[:, overlap_genes].X, dtype="float32")
    
    # Extract spatial transcriptomics data
    ST = stdata[:, overlap_genes]
    
    # Perform mapping to spatial spots
    mapping_matrix = Cell2VisiumSpots(
        S=S,
        ST=ST,
        Gx=Gx1,
        Gy=Gx2,
        Gz=Gx3,
        cluster_matrix=cluster_mapping_matrix,
        device=device
    ).fit(learning_rate=learning_rate, num_epochs=num_epochs, print_each=100)
    return mapping_matrix

def sc2sc(
    scadata,
    stdata_raw,
    mapping_matrix,
    thres=0.1,
    method='max'
):
    """
    Maps single-cell RNA-seq data to spatial cells within defined grids.

    Args:
        scadata (anndata.AnnData): scRNA-seq data.
        stdata_raw (anndata.AnnData): Spatial transcriptomics data with single cell resolution.
        mapping_matrix (np.ndarray): Matrix mapping single cells to spatial spots.
        thres (float, optional): Threshold to filter small values. Default is 0.1.
        method (str, optional): Method to select the cells. Options are 'max' and 'lap'. Default is 'max'.

    Returns:
        anndata.AnnData: scRNA-seq data with spatial information, including similarity scores.
    """
    st_x = []
    st_y = []
    select_ct_index = []
    select_gep = []
    similarity_scores = [] 

    # Determine number of cells to select for each grid
    select_cells_num = np.sum(mapping_matrix > thres, axis=0)
    raw_spot_index = stdata_raw.obs['spot_index'].unique()
    overlap_genes = scadata.uns["overlap"]

    for i in tqdm.tqdm(raw_spot_index):
        # Find index of the current spot
        sorted_spot_indices = np.argsort(raw_spot_index)
        index_of_spot = np.where(raw_spot_index[sorted_spot_indices] == i)[0][0]

        # Extract relevant data for the current spot
        st_temp = stdata_raw[stdata_raw.obs['spot_index'] == i][:, overlap_genes]
        num_cells = select_cells_num[index_of_spot]
        if num_cells == 0:
            num_cells = st_temp.shape[0]

        # Select the top cells based on mapping matrix
        sc_temp = scadata[np.argsort(mapping_matrix[:, index_of_spot], axis=0)[-num_cells:], :][:, overlap_genes]
        cos_result = cos_s(sc_temp.X, st_temp.X).T

        # Select cells based on the chosen method
        if method == 'max':
            select_index = np.argmax(cos_result, axis=1)
            similarity_scores.extend(np.max(cos_result, axis=1).tolist())
        elif method == 'lap':
            import lap
            select_index = lap.lapjv(1 - cos_result, extend_cost=True)[1]
            similarity_scores.extend([cos_result[row, col] for row, col in enumerate(select_index)])
        else:
            raise ValueError("Method must be 'max' or 'lap'.")

        # Collect data for the selected cells
        sc_select = scadata[np.argsort(mapping_matrix[:,index_of_spot], axis=0)[-num_cells:],:][select_index]
        select_ct_index.extend(sc_select.obs.index.tolist())
        select_gep.append(sc_select.X.toarray())
        st_x.extend(st_temp.obs.x.values.tolist())
        st_y.extend(st_temp.obs.y.values.tolist())

    # Create an AnnData object for the result
    cell_alocated_data = sc.AnnData(
        np.vstack(select_gep),
        obs=pd.DataFrame(index=select_ct_index),
        var=scadata.var
    )
    cell_alocated_data.obsm['spatial'] = np.array([st_x, st_y]).T
    cell_alocated_data.obs = scadata[cell_alocated_data.obs.index, :].obs
    cell_alocated_data.obsm['similarity_score'] = np.array(similarity_scores)

    return cell_alocated_data

def sc2st(
    scadata,
    stdata,
    stdata_raw,
    mapping_matrix,
    thres=0.01):
    """
    Mapping the single cells to the single cells of maximized proability spatial spots.
    Args:
        scadata (anndata.AnnData): scRNA-seq data
        stdata_raw (anndata.AnnData): spatial transcriptomics data with single cell resolution.
        mapping_matrix (np.array): sc to spots mapping matrix
        thres (floot): thres to filter small values
        method (str): method to select the cells
    Return:
        cell_alocated_data (anndata.AnnData): scRNA-seq data with spatial information
    """
    overlap_genes = scadata.uns["overlap"]
    scadata = scadata[:,overlap_genes].copy()
    stdata = stdata[:,overlap_genes].copy()
    stdata_raw = stdata_raw[:,overlap_genes].copy()
    stdata_raw.obs['spot_index'] = stdata_raw.obs['spot_index'].astype('str')
    spot_index = stdata[np.argmax(mapping_matrix, axis=1)].obs.index
    raw_spot_index = stdata_raw.obs['spot_index'].unique()
    sc_location = []
    sc_index = []
    import lap
    for i in tqdm.tqdm(raw_spot_index):
        st_temp = stdata_raw[stdata_raw.obs['spot_index'] == i]
        sc_temp = scadata[spot_index == i]
        if len(sc_temp) == 0 :
            continue
        cos_result = cos_s(sc_temp.X,st_temp.X)
        select_index = lap.lapjv(1-cos_result,extend_cost=True)[1]
        sc_index.extend(sc_temp.obs.index.tolist())
        sc_location.extend(st_temp[select_index].obsm['spatial'])
    cell_alocated_data = scadata[sc_index]
    cell_alocated_data.obsm['spatial'] = np.array(sc_location)
    return cell_alocated_data

def main_Polyomino():
    args = argument_parser()
    sc_adata = sc.read_h5ad(args.sc)
    st_adata = sc.read_h5ad(args.st)
    stdata_spot = generate_grid(st_adata, width=args.w)
    if args.custom_region is not None:
        zm_obj = Polyomino(sc_adata, stdata_spot, 
                     cluster_time=args.cluster_time, cluster_thres=args.cluster_thres,custom_label = args.custom_region,
                     device=args.device)
    else:
        zm_obj = Polyomino(sc_adata, stdata_spot, 
                     cluster_time=args.cluster_time, cluster_thres=args.cluster_thres,
                     device=args.device)
    zm_obj.allocate()
    cell_allocated_data = sc2sc(sc_adata, st_adata, zm_obj.spot_matrix, 
                                        thres=args.thres if args.thres else 0.1, 
                                        method=args.method)
    cell_allocated_data.write_h5ad(args.output)
if __name__ == "__main__":
    main_Polyomino()



import argparse

def argument_parser():
    parser = argparse.ArgumentParser(description="CellChIP is an algorithm framework "
                                                 "employing multi-layered regional "
                                                 "constraints to accurately assign "
                                                 "cell locations, enhancing spatial "
                                                 "accuracy and resilience to noise.")

    # Required arguments
    required = parser.add_argument_group('Required')
    required.add_argument("-sc", help="Path of scRNA-Seq anndata", type=str, default=None, required=True)
    required.add_argument("-st", help="Path of spatial anndata", type=str, default=None, required=True)
    required.add_argument("-w", help="width of gridding spatial anndata", type=int, default=None, required=True)

    # Output arguments
    parser.add_argument("-o", "--output", help="Output file", type=str, default="./zmap_assign.h5ad")

    # Custom options
    parser.add_argument("--cluster_time", help="Time allowed for clustering", type=int, default=0)
    parser.add_argument("--custom_region", help="Custom region label for ST data", type=str, default=None)
    parser.add_argument("--cluster_thres", help="Threshold for clustering", type=float, default=None)
    parser.add_argument("--thres", help="Threshold for voxel mapping", type=float, default=0.1)
    parser.add_argument("--method", help="Method used to assing cells to voxels", type=str, choices=['max', 'lap'], default='max')
    parser.add_argument("--device", help="Device to run the computation on", type=str, choices=['cpu', 'cuda'], default='cpu')

    arguments = parser.parse_args()

    return arguments.__dict__



"""
sinkhorn_log - stabilized sinkhorn over log domain with acceleration
[u,v,err] = sinkhorn_log(mu,nu,c,epsilon,options);
mu and nu are marginals.
c is cost
epsilon is regularization
coupling is 
gamma = exp( (-c+u*ones(1,N(2))+ones(N(1),1)*v')/epsilon );

options.niter is the number of iterations.
options.tau is an avering step. 
  - tau=0 is usual sinkhorn
  - tau<0 produces extrapolation and can usually accelerate.
options.rho controls the amount of mass variation. Large value of rho
impose strong constraint on mass conservation. rho=Inf (default)
corresponds to the usual OT (balanced setting). 

Copyright (c) 2016 Gabriel Peyre

@author: linxin,caiqy
"""
import numpy as np
from multiprocessing import Pool
from scipy.linalg.blas import sgemm
def sinkhorn_log(mu, nu, c, epsilon=1e-6, niter=50):
    lamb = 1e4/(1e4+1) # 1e4 is the default value in the original code
    N = np.asarray([mu.size, nu.size])
    H = np.ones((N[0],N[1]))
    u = np.zeros(N[0])
    v = np.zeros(N[1])
    for i in range(niter):
        M = (-c + fast_dot(np.diag(u) , H) + fast_dot(H , np.diag(v)))/epsilon
        u = ave(u, lamb*epsilon*np.log(mu+1e-20) - lamb*epsilon*lse(M) + lamb*u, 1/2)
        v = ave(v, lamb*epsilon*np.log(nu+1e-20) - lamb*epsilon*lse(M.transpose()) + lamb*v, 1/2)
    gamma = np.exp(M)
    return gamma

def ave(u, u1, tau):
    return tau*u+(1-tau)*u1

def lse(A):
    return np.log(np.exp(A).sum(axis=1)+1e-100)

def fast_dot(a,b):
    return sgemm(1, a = a, b = b)

def _solve_ot(k):
    return np.array(sinkhorn_log(scy[k,:], scx[k,:], grid[:,:], 1e-6, 50))

def solve_OT(emptyGrids,scLocX,scLocY,thres=0.01,njob=8):
    """
    This function is used to solve the OT problem.
    The output is a matrix with the same shape as the spatial transcriptomics data, where each element is a probability of the corresponding cell belonging to a spoy.
    """
    # grid需要增加同个权重还是不同权重的区别
    global grid
    grid = emptyGrids
    scLocX_0 = scLocX*(scLocX>thres).astype(int)
    scLocY_0 = scLocY*(scLocY>thres).astype(int)
    zx = scLocX_0.sum(axis=1)
    zy = scLocY_0.sum(axis=1)
    global scx,scy
    scLocX_sel = scLocX[(zx!=0) & (zy!=0),]
    scLocY_sel = scLocY[(zx!=0) & (zy!=0),]
    scx = scLocX_sel * ((scLocX_sel > 0).astype(int))
    scy = scLocY_sel * ((scLocY_sel > 0).astype(int))
    print('Starting solve the OT problem.')
    # multiprocessing
    pool = Pool(njob) 
    gamma = pool.map(_solve_ot, range(scLocX_sel.shape[0]))
    pool.close()
    pool.join()
    gamma = np.array(gamma)
    return gamma


import os,csv,re
import pandas as pd
import numpy as np
import scanpy as sc
from scipy.sparse import issparse
import torch
from sklearn.decomposition import PCA
import math
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
from torch.nn.modules.module import Module
from sklearn.cluster import KMeans
import torch.optim as optim
from random import shuffle
import scipy
from anndata import AnnData,read_csv,read_text,read_mtx
import random
import numba

@numba.njit("f4(f4[:], f4[:])")
def euclid_dist(t1,t2):
	sum=0
	for i in range(t1.shape[0]):
		sum+=(t1[i]-t2[i])**2
	return np.sqrt(sum)

@numba.njit("f4[:,:](f4[:,:])", parallel=True, nogil=True)
def pairwise_distance(X):
	n=X.shape[0]
	adj=np.empty((n, n), dtype=np.float32)
	for i in numba.prange(n):
		for j in numba.prange(n):
			adj[i][j]=euclid_dist(X[i], X[j])
	return adj

def extract_color(x_pixel=None, y_pixel=None, image=None, beta=49):
	#beta to control the range of neighbourhood when calculate grey vale for one spot
	beta_half=round(beta/2)
	g=[]
	for i in range(len(x_pixel)):
		max_x=image.shape[0]
		max_y=image.shape[1]
		nbs=image[max(0,x_pixel[i]-beta_half):min(max_x,x_pixel[i]+beta_half+1),max(0,y_pixel[i]-beta_half):min(max_y,y_pixel[i]+beta_half+1)]
		g.append(np.mean(np.mean(nbs,axis=0),axis=0))
	c0, c1, c2=[], [], []
	for i in g:
		c0.append(i[0])
		c1.append(i[1])
		c2.append(i[2])
	c0=np.array(c0)
	c1=np.array(c1)
	c2=np.array(c2)
	c3=(c0*np.var(c0)+c1*np.var(c1)+c2*np.var(c2))/(np.var(c0)+np.var(c1)+np.var(c2))
	return c3

def calculate_adj_matrix(x, y, x_pixel=None, y_pixel=None, image=None, beta=49, alpha=1, histology=True):
	#x,y,x_pixel, y_pixel are lists
	if histology:
		assert (x_pixel is not None) & (x_pixel is not None) & (image is not None)
		assert (len(x)==len(x_pixel)) & (len(y)==len(y_pixel))
		# print("Calculateing adj matrix using histology image...")
		#beta to control the range of neighbourhood when calculate grey vale for one spot
		#alpha to control the color scale
		beta_half=round(beta/2)
		g=[]
		for i in range(len(x_pixel)):
			max_x=image.shape[0]
			max_y=image.shape[1]
			nbs=image[max(0,x_pixel[i]-beta_half):min(max_x,x_pixel[i]+beta_half+1),max(0,y_pixel[i]-beta_half):min(max_y,y_pixel[i]+beta_half+1)]
			g.append(np.mean(np.mean(nbs,axis=0),axis=0))
		c0, c1, c2=[], [], []
		for i in g:
			c0.append(i[0])
			c1.append(i[1])
			c2.append(i[2])
		c0=np.array(c0)
		c1=np.array(c1)
		c2=np.array(c2)
		print("Var of c0,c1,c2 = ", np.var(c0),np.var(c1),np.var(c2))
		c3=(c0*np.var(c0)+c1*np.var(c1)+c2*np.var(c2))/(np.var(c0)+np.var(c1)+np.var(c2))
		c4=(c3-np.mean(c3))/np.std(c3)
		z_scale=np.max([np.std(x), np.std(y)])*alpha
		z=c4*z_scale
		z=z.tolist()
		print("Var of x,y,z = ", np.var(x),np.var(y),np.var(z))
		X=np.array([x, y, z]).T.astype(np.float32)
	else:
		# print("Calculateing adj matrix using xy only...")
		X=np.array([x, y]).T.astype(np.float32)
	return pairwise_distance(X)
def prefilter_cells(adata,min_counts=None,max_counts=None,min_genes=200,max_genes=None):
    if min_genes is None and min_counts is None and max_genes is None and max_counts is None:
        raise ValueError('Provide one of min_counts, min_genes, max_counts or max_genes.')
    id_tmp=np.asarray([True]*adata.shape[0],dtype=bool)
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,min_genes=min_genes)[0]) if min_genes is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,max_genes=max_genes)[0]) if max_genes is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,min_counts=min_counts)[0]) if min_counts is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,max_counts=max_counts)[0]) if max_counts is not None  else id_tmp
    adata._inplace_subset_obs(id_tmp)
    adata.raw=sc.pp.log1p(adata,copy=True) #check the rowname 
    # print("the var_names of adata.raw: adata.raw.var_names.is_unique=:",adata.raw.var_names.is_unique)
   

def prefilter_genes(adata,min_counts=None,max_counts=None,min_cells=10,max_cells=None):
    if min_cells is None and min_counts is None and max_cells is None and max_counts is None:
        raise ValueError('Provide one of min_counts, min_genes, max_counts or max_genes.')
    id_tmp=np.asarray([True]*adata.shape[1],dtype=bool)
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,min_cells=min_cells)[0]) if min_cells is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,max_cells=max_cells)[0]) if max_cells is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,min_counts=min_counts)[0]) if min_counts is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,max_counts=max_counts)[0]) if max_counts is not None  else id_tmp
    adata._inplace_subset_var(id_tmp)


def prefilter_specialgenes(adata,Gene1Pattern="ERCC",Gene2Pattern="MT-"):
    id_tmp1=np.asarray([not str(name).startswith(Gene1Pattern) for name in adata.var_names],dtype=bool)
    id_tmp2=np.asarray([not str(name).startswith(Gene2Pattern) for name in adata.var_names],dtype=bool)
    id_tmp=np.logical_and(id_tmp1,id_tmp2)
    adata._inplace_subset_var(id_tmp)


def calculate_p(adj, l):
    adj_exp=np.exp(-1*(adj**2)/(2*(l**2)))
    return np.mean(np.sum(adj_exp,1))-1

def test_l(adj, list_l):
    for l in list_l:
        print("l is ",str(l),"Percentage of total expression contributed by neighborhoods:",calculate_p(adj, l))

def find_l(p, adj, start=0.5, end=2,sep=0.01, tol=0.01):
    for l in np.arange(start, end, sep):
        q=calculate_p(adj, l)
        print("L=", str(l), "P=", str(round(q,5)))
        if np.abs(p-q)<=tol:
            return l
    # print("l not found, try bigger range or smaller sep!")

def search_l(p, adj, start=0.01, end=1000, tol=0.01, max_run=100):
    run=0
    p_low=calculate_p(adj, start)
    p_high=calculate_p(adj, end)
    if p_low>p+tol:
        # print("l not found, try smaller start point.")
        return None
    elif p_high<p-tol:
        # print("l not found, try bigger end point.")
        return None
    elif  np.abs(p_low-p) <=tol:
        # print("recommended l = ", str(start))
        return start
    elif  np.abs(p_high-p) <=tol:
        # print("recommended l = ", str(end))
        return end
    while (p_low+tol)<p<(p_high-tol):
        run+=1
        # print("Run "+str(run)+": l ["+str(start)+", "+str(end)+"], p ["+str(p_low)+", "+str(p_high)+"]")
        if run >max_run:
            # print("Exact l not found, closest values are:\n"+"l="+str(start)+": "+"p="+str(p_low)+"\nl="+str(end)+": "+"p="+str(p_high))
            return None
        mid=(start+end)/2
        p_mid=calculate_p(adj, mid)
        if np.abs(p_mid-p)<=tol:
            # print("recommended l = ", str(mid))
            return mid
        if p_mid<=p:
            start=mid
            p_low=p_mid
        else:
            end=mid
            p_high=p_mid

def count_nbr(target_cluster,cell_id, x, y, pred, radius):
    adj_2d=calculate_adj_matrix(x=x,y=y, histology=False)
    cluster_num = dict()
    df = {'cell_id': cell_id, 'x': x, "y":y, "pred":pred}
    df = pd.DataFrame(data=df)
    df.index=df['cell_id']
    target_df=df[df["pred"]==target_cluster]
    row_index=0
    num_nbr=[]
    for index, row in target_df.iterrows():
        x=row["x"]
        y=row["y"]
        tmp_nbr=df[((df["x"]-x)**2+(df["y"]-y)**2)<=(radius**2)]
        num_nbr.append(tmp_nbr.shape[0])
    return np.mean(num_nbr)

def search_radius(target_cluster,cell_id, x, y, pred, start, end, num_min=8, num_max=15,  max_run=100):
    run=0
    num_low=count_nbr(target_cluster,cell_id, x, y, pred, start)
    num_high=count_nbr(target_cluster,cell_id, x, y, pred, end)
    if num_min<=num_low<=num_max:
        # print("recommended radius = ", str(start))
        return start
    elif num_min<=num_high<=num_max:
        # print("recommended radius = ", str(end))
        return end
    elif num_low>num_max:
        # print("Try smaller start.")
        return None
    elif num_high<num_min:
        # print("Try bigger end.")
        return None
    while (num_low<num_min) and (num_high>num_min):
        run+=1
        # print("Run "+str(run)+": radius ["+str(start)+", "+str(end)+"], num_nbr ["+str(num_low)+", "+str(num_high)+"]")
        if run >max_run:
            # print("Exact radius not found, closest values are:\n"+"radius="+str(start)+": "+"num_nbr="+str(num_low)+"\nradius="+str(end)+": "+"num_nbr="+str(num_high))
            return None
        mid=(start+end)/2
        num_mid=count_nbr(target_cluster,cell_id, x, y, pred, mid)
        if num_min<=num_mid<=num_max:
            # print("recommended radius = ", str(mid), "num_nbr="+str(num_mid))
            return mid
        if num_mid<num_min:
            start=mid
            num_low=num_mid
        elif num_mid>num_max:
            end=mid
            num_high=num_mid

def find_neighbor_clusters(target_cluster,cell_id, x, y, pred,radius, ratio=1/2):
    cluster_num = dict()
    for i in pred:
        cluster_num[i] = cluster_num.get(i, 0) + 1
    df = {'cell_id': cell_id, 'x': x, "y":y, "pred":pred}
    df = pd.DataFrame(data=df)
    df.index=df['cell_id']
    target_df=df[df["pred"]==target_cluster]
    nbr_num={}
    row_index=0
    num_nbr=[]
    for index, row in target_df.iterrows():
        x=row["x"]
        y=row["y"]
        tmp_nbr=df[((df["x"]-x)**2+(df["y"]-y)**2)<=(radius**2)]
        #tmp_nbr=df[(df["x"]<x+radius) & (df["x"]>x-radius) & (df["y"]<y+radius) & (df["y"]>y-radius)]
        num_nbr.append(tmp_nbr.shape[0])
        for p in tmp_nbr["pred"]:
            nbr_num[p]=nbr_num.get(p,0)+1
    del nbr_num[target_cluster]
    nbr_num_back=nbr_num.copy() #Backup
    nbr_num=[(k, v)  for k, v in nbr_num.items() if v>(ratio*cluster_num[k])]
    nbr_num.sort(key=lambda x: -x[1])
    # print("radius=", radius, "average number of neighbors for each spot is", np.mean(num_nbr))
    # print(" Cluster",target_cluster, "has neighbors:")
    for t in nbr_num:
        print("Dmain ", t[0], ": ",t[1])
    ret=[t[0] for t in nbr_num]
    if len(ret)==0:
        nbr_num_back=[(k, v)  for k, v in nbr_num_back.items()]
        nbr_num_back.sort(key=lambda x: -x[1])
        ret=[nbr_num_back[0][0]]
        # print("No neighbor domain found, only return one potential neighbor domain:",ret)
        # print("Try bigger radius or smaller ratio.")
    return ret


def rank_genes_groups(input_adata, target_cluster,nbr_list, label_col, adj_nbr=True, log=False):
    if adj_nbr:
        nbr_list=nbr_list+[target_cluster]
        adata=input_adata[input_adata.obs[label_col].isin(nbr_list)]
    else:
        adata=input_adata.copy()
    adata.var_names_make_unique()
    adata.obs["target"]=((adata.obs[label_col]==target_cluster)*1).astype('category')
    sc.tl.rank_genes_groups(adata, groupby="target",reference="rest", n_genes=adata.shape[1],method='wilcoxon')
    pvals_adj=[i[0] for i in adata.uns['rank_genes_groups']["pvals_adj"]]
    genes=[i[1] for i in adata.uns['rank_genes_groups']["names"]]
    if issparse(adata.X):
        obs_tidy=pd.DataFrame(adata.X.A)
    else:
        obs_tidy=pd.DataFrame(adata.X)
    obs_tidy.index=adata.obs["target"].tolist()
    obs_tidy.columns=adata.var.index.tolist()
    obs_tidy=obs_tidy.loc[:,genes]
    # 1. compute mean value
    mean_obs = obs_tidy.groupby(level=0).mean()
    # 2. compute fraction of cells having value >0
    obs_bool = obs_tidy.astype(bool)
    fraction_obs = obs_bool.groupby(level=0).sum() / obs_bool.groupby(level=0).count()
    # compute fold change.
    if log: #The adata already logged
        fold_change=np.exp((mean_obs.loc[1] - mean_obs.loc[0]).values)
    else:
        fold_change = (mean_obs.loc[1] / (mean_obs.loc[0]+ 1e-9)).values
    df = {'genes': genes, 'in_group_fraction': fraction_obs.loc[1].tolist(), "out_group_fraction":fraction_obs.loc[0].tolist(),"in_out_group_ratio":(fraction_obs.loc[1]/fraction_obs.loc[0]).tolist(),"in_group_mean_exp": mean_obs.loc[1].tolist(), "out_group_mean_exp": mean_obs.loc[0].tolist(),"fold_change":fold_change.tolist(), "pvals_adj":pvals_adj}
    df = pd.DataFrame(data=df)
    return df

def relative_func(expres):
    #expres: an array counts expression for a gene
    maxd = np.max(expres) - np.min(expres)
    min_exp=np.min(expres)
    rexpr = (expres - min_exp)/maxd
    return rexpr

def plot_relative_exp(input_adata, gene, x_name, y_name,color,use_raw=False, spot_size=200000):
    adata=input_adata.copy()
    if use_raw:
        X=adata.raw.X
    else:
        X=adata.X
    if issparse(X):
        X=pd.DataFrame(X.A)
    else:
        X=pd.DataFrame(X)
    X.index=adata.obs.index
    X.columns=adata.var.index
    rexpr=relative_func(X.loc[:,gene])
    adata.obs["rexpr"]=rexpr
    fig=sc.pl.scatter(adata,x=x_name,y=y_name,color="rexpr",title=gene+"_rexpr",color_map=color,show=False,size=spot_size/adata.shape[0])
    return fig

def plot_log_exp(input_adata, gene, x_name, y_name,color,use_raw=False):
    adata=input_adata.copy()
    if use_raw:
        X=adata.X
    else:
        X=adata.raw.X
    if issparse(X):
        X=pd.DataFrame(X.A)
    else:
        X=pd.DataFrame(X)
    X.index=adata.obs.index
    X.columns=adata.var.index
    adata.obs["log"]=np.log((X.loc[:,gene]+1).tolist())
    fig=sc.pl.scatter(adata,x=x_name,y=y_name,color="log",title=gene+"_log",color_map=color,show=False,size=200000/adata.shape[0])
    return fig

def detect_subclusters(cell_id, x, y, pred, target_cluster, radius=3, res=0.2):
    df = {'cell_id': cell_id, 'x': x, "y":y, "pred":pred}
    df = pd.DataFrame(data=df)
    df.index=df['cell_id']
    target_df=df[df["pred"]==target_cluster]
    nbr=np.zeros([target_df.shape[0],len(set(df["pred"]))],dtype=int)
    num_nbr=[]
    row_index=0
    for index, row in target_df.iterrows():
        x=row["x"]
        y=row["y"]
        tmp_nbr=df[(df["x"]<x+radius) & (df["x"]>x-radius) & (df["y"]<y+radius) & (df["y"]>y-radius)]
        num_nbr.append(tmp_nbr.shape[0])
        for p in tmp_nbr["pred"]:
            nbr[row_index,int(p)]+=1
        row_index+=1
    #Minus out the cell itself
    nbr[:,target_cluster]=nbr[:,target_cluster]-1
    nbr=sc.AnnData(nbr)
    sc.pp.neighbors(nbr, n_neighbors=10)
    sc.tl.leiden(nbr,resolution=res)
    sub_cluster=nbr.obs['leiden'].astype(int).to_numpy()
    target_df["sub_cluster"]=sub_cluster
    target_df["sub_cluster"]=target_df["sub_cluster"].astype('category')
    tmp=[]
    for j in df.index:
        if j in target_df.index:
            tmp.append(target_df.loc[j,"sub_cluster"])
        else:
            tmp.append("-1")
    #ret = {'cell_id': cell_id, 'sub_cluster_'+str(target_cluster): tmp}
    #ret = pd.DataFrame(data=ret)
    #ret.index=ret['cell_id']
    ret=tmp
    return ret

def find_meta_gene(input_adata,
                    pred,
                    target_domain,
                    start_gene,
                    mean_diff=0,
                    early_stop=True,
                    max_iter=5,
                    use_raw=False):
    meta_name=start_gene
    adata=input_adata.copy()
    adata.obs["meta"]=adata.X[:,adata.var.index==start_gene]
    adata.obs["pred"]=pred
    num_non_target=adata.shape[0]
    for i in range(max_iter):
        #Select cells
        tmp=adata[((adata.obs["meta"]>np.mean(adata.obs[adata.obs["pred"]==target_domain]["meta"]))|(adata.obs["pred"]==target_domain))]
        tmp.obs["target"]=((tmp.obs["pred"]==target_domain)*1).astype('category').copy()
        if (len(set(tmp.obs["target"]))<2) or (np.min(tmp.obs["target"].value_counts().values)<5):
            # print("Meta gene is: ", meta_name)
            return meta_name, adata.obs["meta"].tolist()
        #DE
        sc.tl.rank_genes_groups(tmp, groupby="target",reference="rest", n_genes=1,method='wilcoxon')
        adj_g=tmp.uns['rank_genes_groups']["names"][0][0]
        add_g=tmp.uns['rank_genes_groups']["names"][0][1]
        meta_name_cur=meta_name+"+"+add_g+"-"+adj_g
        # print("Add gene: ", add_g)
        # print("Minus gene: ", adj_g)
        #Meta gene
        adata.obs[add_g]=adata.X[:,adata.var.index==add_g]
        adata.obs[adj_g]=adata.X[:,adata.var.index==adj_g]
        adata.obs["meta_cur"]=(adata.obs["meta"]+adata.obs[add_g]-adata.obs[adj_g])
        adata.obs["meta_cur"]=adata.obs["meta_cur"]-np.min(adata.obs["meta_cur"])
        mean_diff_cur=np.mean(adata.obs["meta_cur"][adata.obs["pred"]==target_domain])-np.mean(adata.obs["meta_cur"][adata.obs["pred"]!=target_domain])
        num_non_target_cur=np.sum(tmp.obs["target"]==0)
        if (early_stop==False) | ((num_non_target>=num_non_target_cur) & (mean_diff<=mean_diff_cur)):
            num_non_target=num_non_target_cur
            mean_diff=mean_diff_cur
            # print("Absolute mean change:", mean_diff)
            # print("Number of non-target spots reduced to:",num_non_target)
        else:
            # print("Stopped!", "Previous Number of non-target spots",num_non_target, num_non_target_cur, mean_diff,mean_diff_cur)
            # print("Previous Number of non-target spots",num_non_target, num_non_target_cur, mean_diff,mean_diff_cur)
            # print("Previous Number of non-target spots",num_non_target)
            # print("Current Number of non-target spots",num_non_target_cur)
            # print("Absolute mean change", mean_diff)
            # print("===========================================================================")
            # print("Meta gene: ", meta_name)
            # print("===========================================================================")
            return meta_name, adata.obs["meta"].tolist()
        meta_name=meta_name_cur
        adata.obs["meta"]=adata.obs["meta_cur"]
        # print("===========================================================================")
        # print("Meta gene is: ", meta_name)
        # print("===========================================================================")
    return meta_name, adata.obs["meta"].tolist()


def search_res(adata, adj,embed, l, target_num, start=0.4, step=0.1, tol=5e-3, lr=0.05, max_epochs=10, r_seed=100, t_seed=100, n_seed=100, max_run=10):
    random.seed(r_seed)
    torch.manual_seed(t_seed)
    np.random.seed(n_seed)
    res=start
    # print("Start at res = ", res, "step = ", step)
    clf=SpaGCN()
    clf.set_l(l)
    clf.train(adata,adj,embed,init_spa=True,init="leiden",res=res, tol=tol, lr=lr, max_epochs=max_epochs)
    y_pred, _=clf.predict()
    old_num=len(set(y_pred))
    # print("Res = ", res, "Num of clusters = ", old_num)
    run=0
    while old_num!=target_num:
        random.seed(r_seed)
        torch.manual_seed(t_seed)
        np.random.seed(n_seed)
        old_sign=1 if (old_num<target_num) else -1
        clf=SpaGCN()
        clf.set_l(l)
        clf.train(adata,adj,embed,init_spa=True,init="leiden",res=res+step*old_sign, tol=tol, lr=lr, max_epochs=max_epochs)
        y_pred, _=clf.predict()
        new_num=len(set(y_pred))
        # print("Res = ", res+step*old_sign, "Num of clusters = ", new_num)
        if new_num==target_num:
            res=res+step*old_sign
            # print("recommended res = ", str(res))
            return res
        new_sign=1 if (new_num<target_num) else -1
        if new_sign==old_sign:
            res=res+step*old_sign
            # print("Res changed to", res)
            old_num=new_num
        else:
            step=step/2
            # print("Step changed to", step)
        if run >max_run:
            # print("Exact resolution not found")
            # print("Recommended res = ", str(res))
            return res
        run+=1
    # print("recommended res = ", str(res))
    return res


def refine(sample_id, pred, dis, shape="hexagon"):
    refined_pred=[]
    pred=pd.DataFrame({"pred": pred}, index=sample_id)
    dis_df=pd.DataFrame(dis, index=sample_id, columns=sample_id)
    if shape=="hexagon":
        num_nbs=6 
    elif shape=="square":
        num_nbs=4
    else:
        print("Shape not recongized, shape='hexagon' for Visium data, 'square' for ST data.")
    for i in range(len(sample_id)):
        index=sample_id[i]
        dis_tmp=dis_df.loc[index, :].sort_values()
        nbs=dis_tmp[0:num_nbs+1]
        nbs_pred=pred.loc[nbs.index, "pred"]
        self_pred=pred.loc[index, "pred"]
        v_c=nbs_pred.value_counts()
        if (v_c.loc[self_pred]<num_nbs/2) and (np.max(v_c)>num_nbs/2):
            refined_pred.append(v_c.idxmax())
        else:           
            refined_pred.append(self_pred)
    return refined_pred
class GraphConvolution(Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """

    def __init__(self, in_features, out_features, bias=True):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.FloatTensor(in_features, out_features))
        if bias:
            self.bias = Parameter(torch.FloatTensor(out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, input, adj):
        support = torch.mm(input, self.weight)
        output = torch.spmm(adj, support)
        if self.bias is not None:
            return output + self.bias
        else:
            return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'
class simple_GC_DEC(nn.Module):
    def __init__(self, nfeat, nhid, alpha=0.2):
        super(simple_GC_DEC, self).__init__()
        self.gc = GraphConvolution(nfeat, nhid)
        self.nhid=nhid
        #self.mu determined by the init method
        self.alpha=alpha

    def forward(self, x, adj):
        x=self.gc(x, adj)
        q = 1.0 / ((1.0 + torch.sum((x.unsqueeze(1) - self.mu)**2, dim=2) / self.alpha) + 1e-8)
        q = q**(self.alpha+1.0)/2.0
        q = q / torch.sum(q, dim=1, keepdim=True)
        return x, q

    def loss_function(self, p, q):
        def kld(target, pred):
            return torch.mean(torch.sum(target*torch.log(target/(pred+1e-6)), dim=1))
        loss = kld(p, q)
        return loss

    def target_distribution(self, q):
        #weight = q ** 2 / q.sum(0)
        #return torch.transpose((torch.transpose(weight,0,1) / weight.sum(1)),0,1)e
        p = q**2 / torch.sum(q, dim=0)
        p = p / torch.sum(p, dim=1, keepdim=True)
        return p

    def fit(self, X,adj,  lr=0.001, max_epochs=5000, update_interval=3, trajectory_interval=50,weight_decay=5e-4,opt="sgd",init="leiden",n_neighbors=10,res=0.4,n_clusters=10,init_spa=True,tol=1e-3):
        self.trajectory=[]
        if opt=="sgd":
            optimizer = optim.SGD(self.parameters(), lr=lr, momentum=0.9)
        elif opt=="admin":
            optimizer = optim.Adam(self.parameters(),lr=lr, weight_decay=weight_decay)

        features= self.gc(torch.FloatTensor(X),torch.FloatTensor(adj))
        #----------------------------------------------------------------        
        if init=="kmeans":
            # print("Initializing cluster centers with kmeans, n_clusters known")
            self.n_clusters=n_clusters
            kmeans = KMeans(self.n_clusters, n_init=20)
            if init_spa:
                #------Kmeans use exp and spatial
                y_pred = kmeans.fit_predict(features.detach().numpy())
            else:
                #------Kmeans only use exp info, no spatial
                y_pred = kmeans.fit_predict(X)  #Here we use X as numpy
        elif init=="leiden":
            # print("Initializing cluster centers with leiden, resolution = ", res)
            if init_spa:
                adata=sc.AnnData(features.detach().numpy())
            else:
                adata=sc.AnnData(X)
            sc.pp.neighbors(adata, n_neighbors=n_neighbors)
            sc.tl.leiden(adata,resolution=res)
            y_pred=adata.obs['leiden'].astype(int).to_numpy()
            self.n_clusters=len(np.unique(y_pred))
        #----------------------------------------------------------------
        y_pred_last = y_pred
        self.mu = Parameter(torch.Tensor(self.n_clusters, self.nhid))
        X=torch.FloatTensor(X)
        adj=torch.FloatTensor(adj)
        self.trajectory.append(y_pred)
        features=pd.DataFrame(features.detach().numpy(),index=np.arange(0,features.shape[0]))
        Group=pd.Series(y_pred,index=np.arange(0,features.shape[0]),name="Group")
        Mergefeature=pd.concat([features,Group],axis=1)
        cluster_centers=np.asarray(Mergefeature.groupby("Group").mean())
        
        self.mu.data.copy_(torch.Tensor(cluster_centers))
        self.train()
        for epoch in range(max_epochs):
            if epoch%update_interval == 0:
                _, q = self.forward(X,adj)
                p = self.target_distribution(q).data
            # if epoch%10==0:
            #     print("Epoch ", epoch) 
            optimizer.zero_grad()
            z,q = self(X, adj)
            loss = self.loss_function(p, q)
            loss.backward()
            optimizer.step()
            if epoch%trajectory_interval == 0:
                self.trajectory.append(torch.argmax(q, dim=1).data.cpu().numpy())

            #Check stop criterion
            y_pred = torch.argmax(q, dim=1).data.cpu().numpy()
            delta_label = np.sum(y_pred != y_pred_last).astype(np.float32) / X.shape[0]
            y_pred_last = y_pred
            if epoch>0 and (epoch-1)%update_interval == 0 and delta_label < tol:
                # print('delta_label ', delta_label, '< tol ', tol)
                # print("Reach tolerance threshold. Stopping training.")
                # print("Total epoch:", epoch)
                break


    def fit_with_init(self, X,adj, init_y, lr=0.001, max_epochs=5000, update_interval=1, weight_decay=5e-4,opt="sgd"):
        # print("Initializing cluster centers with kmeans.")
        if opt=="sgd":
            optimizer = optim.SGD(self.parameters(), lr=lr, momentum=0.9)
        elif opt=="admin":
            optimizer = optim.Adam(self.parameters(),lr=lr, weight_decay=weight_decay)
        X=torch.FloatTensor(X)
        adj=torch.FloatTensor(adj)
        features, _ = self.forward(X,adj)
        features=pd.DataFrame(features.detach().numpy(),index=np.arange(0,features.shape[0]))
        Group=pd.Series(init_y,index=np.arange(0,features.shape[0]),name="Group")
        Mergefeature=pd.concat([features,Group],axis=1)
        cluster_centers=np.asarray(Mergefeature.groupby("Group").mean())
        self.mu.data.copy_(torch.Tensor(cluster_centers))
        self.train()
        for epoch in range(max_epochs):
            if epoch%update_interval == 0:
                _, q = self.forward(torch.FloatTensor(X),torch.FloatTensor(adj))
                p = self.target_distribution(q).data
            X=torch.FloatTensor(X)
            adj=torch.FloatTensor(adj)
            optimizer.zero_grad()
            z,q = self(X, adj)
            loss = self.loss_function(p, q)
            loss.backward()
            optimizer.step()

    def predict(self, X, adj):
        z,q = self(torch.FloatTensor(X),torch.FloatTensor(adj))
        return z, q




class GC_DEC(nn.Module):
    def __init__(self, nfeat, nhid1,nhid2, n_clusters=None, dropout=0.5,alpha=0.2):
        super(GC_DEC, self).__init__()

        self.gc1 = GraphConvolution(nfeat, nhid1)
        self.gc2 = GraphConvolution(nhid1, nhid2)
        self.dropout = dropout
        self.mu = Parameter(torch.Tensor(n_clusters, nhid2))
        self.n_clusters=n_clusters
        self.alpha=alpha

    def forward(self, x, adj):
        x=self.gc1(x, adj)
        x = F.relu(x)
        x = F.dropout(x, self.dropout, training=True)
        x = self.gc2(x, adj)
        q = 1.0 / ((1.0 + torch.sum((x.unsqueeze(1) - self.mu)**2, dim=2) / self.alpha) + 1e-6)
        q = q**(self.alpha+1.0)/2.0
        q = q / torch.sum(q, dim=1, keepdim=True)
        return x, q

    def loss_function(self, p, q):
        def kld(target, pred):
            return torch.mean(torch.sum(target*torch.log(target/(pred+1e-6)), dim=1))
        loss = kld(p, q)
        return loss

    def target_distribution(self, q):
        #weight = q ** 2 / q.sum(0)
        #return torch.transpose((torch.transpose(weight,0,1) / weight.sum(1)),0,1)e
        p = q**2 / torch.sum(q, dim=0)
        p = p / torch.sum(p, dim=1, keepdim=True)
        return p

    def fit(self, X,adj, lr=0.001, max_epochs=10, update_interval=5, weight_decay=5e-4,opt="sgd",init="leiden",n_neighbors=10,res=0.4):
        self.trajectory=[]
        # print("Initializing cluster centers with kmeans.")
        if opt=="sgd":
            optimizer = optim.SGD(self.parameters(), lr=lr, momentum=0.9)
        elif opt=="admin":
            optimizer = optim.Adam(self.parameters(),lr=lr, weight_decay=weight_decay)

        features, _ = self.forward(torch.FloatTensor(X),torch.FloatTensor(adj))
        #----------------------------------------------------------------
        
        if init=="kmeans":
            #Kmeans only use exp info, no spatial
            #kmeans = KMeans(self.n_clusters, n_init=20)
            #y_pred = kmeans.fit_predict(X)  #Here we use X as numpy
            #Kmeans use exp and spatial
            kmeans = KMeans(self.n_clusters, n_init=20)
            y_pred = kmeans.fit_predict(features.detach().numpy())
        elif init=="leiden":
            adata=sc.AnnData(features.detach().numpy())
            sc.pp.neighbors(adata, n_neighbors=n_neighbors)
            sc.tl.leiden(adata,resolution=res)
            y_pred=adata.obs['leiden'].astype(int).to_numpy()
        #----------------------------------------------------------------
        X=torch.FloatTensor(X)
        adj=torch.FloatTensor(adj)
        self.trajectory.append(y_pred)
        features=pd.DataFrame(features.detach().numpy(),index=np.arange(0,features.shape[0]))
        Group=pd.Series(y_pred,index=np.arange(0,features.shape[0]),name="Group")
        Mergefeature=pd.concat([features,Group],axis=1)
        cluster_centers=np.asarray(Mergefeature.groupby("Group").mean())
        
        self.mu.data.copy_(torch.Tensor(cluster_centers))
        self.train()
        for epoch in range(max_epochs):
            if epoch%update_interval == 0:
                _, q = self.forward(X,adj)
                p = self.target_distribution(q).data
            if epoch%100==0:
                print("Epoch ", epoch) 
            optimizer.zero_grad()
            z,q = self(X, adj)
            loss = self.loss_function(p, q)
            loss.backward()
            optimizer.step()
            self.trajectory.append(torch.argmax(q, dim=1).data.cpu().numpy())

    def fit_with_init(self, X,adj, init_y, lr=0.001, max_epochs=10, update_interval=1, weight_decay=5e-4,opt="sgd"):
        # print("Initializing cluster centers with kmeans.")
        if opt=="sgd":
            optimizer = optim.SGD(self.parameters(), lr=lr, momentum=0.9)
        elif opt=="admin":
            optimizer = optim.Adam(self.parameters(),lr=lr, weight_decay=weight_decay)
        X=torch.FloatTensor(X)
        adj=torch.FloatTensor(adj)
        features, _ = self.forward(X,adj)
        features=pd.DataFrame(features.detach().numpy(),index=np.arange(0,features.shape[0]))
        Group=pd.Series(init_y,index=np.arange(0,features.shape[0]),name="Group")
        Mergefeature=pd.concat([features,Group],axis=1)
        cluster_centers=np.asarray(Mergefeature.groupby("Group").mean())
        self.mu.data.copy_(torch.Tensor(cluster_centers))
        self.train()
        for epoch in range(max_epochs):
            if epoch%update_interval == 0:
                _, q = self.forward(torch.FloatTensor(X),torch.FloatTensor(adj))
                p = self.target_distribution(q).data
            X=torch.FloatTensor(X)
            adj=torch.FloatTensor(adj)
            optimizer.zero_grad()
            z,q = self(X, adj)
            loss = self.loss_function(p, q)
            loss.backward()
            optimizer.step()

    def predict(self, X, adj):
        z,q = self(torch.FloatTensor(X),torch.FloatTensor(adj))
        return z, q

class SpaGCN(object):
    def __init__(self):
        super(SpaGCN, self).__init__()
        self.l=None

    def set_l(self, l):
        self.l=l

    def train(self,adata,adj, embed,
            num_pcs=50, 
            lr=0.005,
            max_epochs=2000,
            weight_decay=0,
            opt="admin",
            init_spa=True,
            init="leiden", #leiden or kmeans
            n_neighbors=10, #for leiden
            n_clusters=None, #for kmeans
            res=0.4, #for leiden
            tol=1e-3):
        self.num_pcs=num_pcs
        self.res=res
        self.lr=lr
        self.max_epochs=max_epochs
        self.weight_decay=weight_decay
        self.opt=opt
        self.init_spa=init_spa
        self.init=init
        self.n_neighbors=n_neighbors
        self.n_clusters=n_clusters
        self.res=res
        self.tol=tol
        assert adata.shape[0]==adj.shape[0]==adj.shape[1]
        # pca = PCA(n_components=self.num_pcs)
        # if issparse(adata.X):
        #     pca.fit(adata.X.A)
        #     embed=pca.transform(adata.X.A)
        # else:
        #     pca.fit(adata.X)
        #     embed=pca.transform(adata.X)
        ###------------------------------------------###
        if self.l is None:
            raise ValueError('l should not be set before fitting the model!')
        adj_exp=np.exp(-1*(adj**2)/(2*(self.l**2)))
        #----------Train model----------
        self.model=simple_GC_DEC(embed.shape[1],embed.shape[1])
        self.model.fit(embed,adj_exp,lr=self.lr,max_epochs=self.max_epochs,weight_decay=self.weight_decay,opt=self.opt,init_spa=self.init_spa,init=self.init,n_neighbors=self.n_neighbors,n_clusters=self.n_clusters,res=self.res, tol=self.tol)
        self.embed=embed
        self.adj_exp=adj_exp

    def predict(self):
        z,q=self.model.predict(self.embed,self.adj_exp)
        y_pred = torch.argmax(q, dim=1).data.cpu().numpy()
        # Max probability plot
        prob=q.detach().numpy()
        return y_pred, prob


import numpy as np
import pandas as pd
import torch
from torch.nn.functional import softmax, cosine_similarity

class Cell2Clusters:
    """
    Maps each cell to its region based on pseudo spatial expression matrix and reconstructed cell expression vectors.
    
    Attributes:
        scdata (torch.Tensor): Single-cell expression matrix.
        clusters (torch.Tensor): Spatial cluster expression matrix.
        lambda_g1 (float): Weight for the first term in the loss function.
        lambda_g2 (float): Weight for the second term in the loss function.
        device (str): Device to use ('cpu' or 'cuda').
        random_state (int): Random seed for initialization.
    """
    
    def __init__(
        self,
        scdata,
        clusters,
        lambda_g1=1.0,
        lambda_g2=1.0,
        device="cpu",
        random_state=1597,
    ):
        """
        Initializes the Cell2Clusters instance.
        
        Args:
            scdata (ndarray): Single-cell expression matrix.
            clusters (ndarray): Regional spatial cluster expression matrix.
            lambda_g1 (float): Weight for the first term in the loss function.
            lambda_g2 (float): Weight for the second term in the loss function.
            device (str): Device to use ('cpu' or 'cuda').
            random_state (int): Random seed for initialization.
        """
        self.scdata = torch.tensor(scdata, device=device, dtype=torch.float32)
        self.clusters = torch.tensor(clusters, device=device, dtype=torch.float32)
        self.lambda_g1 = lambda_g1
        self.lambda_g2 = lambda_g2
        self.device = device
        self.random_state = random_state
        np.random.seed(seed=self.random_state)
        self.cluster_matrix = torch.tensor(
            np.random.normal(0, 1, (scdata.shape[0], clusters.shape[0])),
            device=device, requires_grad=True, dtype=torch.float32
        )

    def _loss_fn(self, verbose=False):
        """
        Computes the loss function for the cell-to-cluster mapping.
        
        Args:
            verbose (bool): If True, prints the loss.
        
        Returns:
            tuple: Total loss, cluster cosine similarity (term 1), and cluster cosine similarity (term 2).
        """
        cluster_probs = softmax(self.cluster_matrix, dim=1)
        G_rec = torch.matmul(cluster_probs.t(), self.scdata)
        cluster_cos1 = self.lambda_g1 * cosine_similarity(G_rec, self.clusters, dim=0).mean()
        cluster_cos2 = self.lambda_g2 * cosine_similarity(G_rec, self.clusters, dim=1).mean()
        expression_term = cluster_cos1 + cluster_cos2
        total_loss = -expression_term
        
        if verbose:
            print(f"Total loss: {total_loss.item():.3f}")
        
        return total_loss, cluster_cos1, cluster_cos2

    def fit(self, num_epochs = 500,learning_rate=0.1, print_each=100):
        """
        Runs the optimization process and returns the mapping matrix.
        
        Args:
            num_epochs (int): Number of epochs to train.
            learning_rate (float): Learning rate for the optimizer.
            print_each (int): Frequency of loss printing.
        
        Returns:
            ndarray: Optimized cluster mapping matrix (number_cells x number_clusters).
        """
        torch.manual_seed(seed=self.random_state)
        optimizer = torch.optim.Adam([self.cluster_matrix], lr=learning_rate)
        
        for t in range(num_epochs):
            loss, _, _ = self._loss_fn(verbose=(print_each is not None and t % print_each == 0))
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        
        with torch.no_grad():
            cluster_probs = softmax(self.cluster_matrix, dim=1).cpu().numpy()
        
        # Clear CUDA memory if using GPU
        if self.device == 'cuda':
            torch.cuda.empty_cache()
        
        return cluster_probs

class Cell2Spots:
    """
    Maps each cell to its spot based on multi layer regional spatial constraints and reconstructed cell expression vectors.
    
    Attributes:
        S (torch.Tensor): Single-cell expression matrix.
        ST (AnnData): Spatial data.
        Gx (torch.Tensor): X-axis spatial expression matrix.
        Gy (torch.Tensor): Y-axis spatial expression matrix.
        lambda_gx1 (float): Weight for the first term in X-axis loss function.
        lambda_gx2 (float): Weight for the second term in X-axis loss function.
        lambda_gy1 (float): Weight for the first term in Y-axis loss function.
        lambda_gy2 (float): Weight for the second term in Y-axis loss function.
        cluster_matrix (torch.Tensor): Initial cluster matrix for optimization.
        device (str): Device to use ('cpu' or 'cuda').
        random_state (int): Random seed for initialization.
    """
    
    def __init__(
        self,
        S,
        ST,
        Gx,
        Gy,
        lambda_gx1=1,
        lambda_gx2=0,
        lambda_gy1=1,
        lambda_gy2=0,
        cluster_matrix=None,
        device="cpu",
        random_state=1597,
    ):
        """
        Initializes the Cell2Spots instance.
        
        Args:
            S (ndarray): Single-cell expression matrix.
            ST (AnnData): Spatial data.
            Gx (ndarray): X-axis spatial expression matrix.
            Gy (ndarray): Y-axis spatial expression matrix.
            lambda_gx1 (float): Weight for the first term in X-axis loss function.
            lambda_gx2 (float): Weight for the second term in X-axis loss function.
            lambda_gy1 (float): Weight for the first term in Y-axis loss function.
            lambda_gy2 (float): Weight for the second term in Y-axis loss function.
            cluster_matrix (ndarray or None): Initial cluster matrix for optimization.
            device (str): Device to use ('cpu' or 'cuda').
            random_state (int): Random seed for initialization.
        """
        self.S = torch.tensor(S, device=device, dtype=torch.float32)
        self.ST = ST
        self.Gx = torch.tensor(Gx, device=device, dtype=torch.float32)
        self.Gy = torch.tensor(Gy, device=device, dtype=torch.float32)
        self.lambda_gx1 = lambda_gx1
        self.lambda_gx2 = lambda_gx2
        self.lambda_gy1 = lambda_gy1
        self.lambda_gy2 = lambda_gy2
        self.random_state = random_state
        np.random.seed(seed=self.random_state)
        
        if cluster_matrix is None:
            self.cluster_matrix = torch.tensor(
                np.random.normal(0, 1, (S.shape[0], ST.shape[0])), 
                device=device, requires_grad=True, dtype=torch.float32
            )
            self.mask = None
        else:
            self.cluster_matrix = torch.tensor(
                cluster_matrix, device=device, requires_grad=True, dtype=torch.float32
            )
            self.mask = self.cluster_matrix.clone().detach() < 1
        
        self.device = device
        self.x_length = int(ST.obs['x'].max() - ST.obs['x'].min() + 1)
        self.y_length = int(ST.obs['y'].max() - ST.obs['y'].min() + 1)
        self.x_index = torch.tensor(
            ST.obs['x'] - ST.obs['x'].min(), device=device, dtype=torch.int32
        )
        self.y_index = torch.tensor(
            ST.obs['y'] - ST.obs['y'].min(), device=device, dtype=torch.int32
        )

    def _generate_Xstrips(self, cluster_probs):
        """
        Generate X matrix based on cluster probabilities.
        
        Args:
            cluster_probs (torch.Tensor): Cluster probabilities.
        
        Returns:
            torch.Tensor: X matrix.
        """
        mapping_x = torch.zeros((self.x_length, self.S.shape[0]), device=self.device)
        mapping_x.index_add_(0, self.x_index, cluster_probs.T)
        return mapping_x

    def _generate_Ystrips(self, cluster_probs):
        """
        Generate Y matrix based on cluster probabilities.
        
        Args:
            cluster_probs (torch.Tensor): Cluster probabilities.
        
        Returns:
            torch.Tensor: Y matrix.
        """
        mapping_y = torch.zeros((self.y_length, self.S.shape[0]), device=self.device)
        mapping_y.index_add_(0, self.y_index, cluster_probs.T)
        return mapping_y
    
    def _loss_fn(self, verbose=True):
        """
        Computes the loss function for the cell-to-spot mapping.
        
        Args:
            verbose (bool): If True, prints the loss.
        
        Returns:
            tuple: Total loss, X-axis main loss terms, Y-axis main loss terms, and mask sum.
        """
        cluster_probs = softmax(self.cluster_matrix, dim=1)
        mask_sum = torch.masked_select(cluster_probs, self.mask).mean() if self.mask is not None else 0
        
        Mx = self._generate_Xstrips(cluster_probs)
        My = self._generate_Ystrips(cluster_probs)
        
        Gx_pred = torch.matmul(Mx, self.S)
        Gy_pred = torch.matmul(My, self.S)
        
        gx1 = self.lambda_gx1 * cosine_similarity(Gx_pred, self.Gx, dim=0).mean()
        gx2 = self.lambda_gx2 * cosine_similarity(Gx_pred, self.Gx, dim=1).mean()
        gy1 = self.lambda_gy1 * cosine_similarity(Gy_pred, self.Gy, dim=0).mean()
        gy2 = self.lambda_gy2 * cosine_similarity(Gy_pred, self.Gy, dim=1).mean()
        
        cos_x = gx1 + gx2
        cos_y = gy1 + gy2
        
        total_loss = -cos_x - cos_y + mask_sum
        
        if verbose:
            print(f"Total loss: {total_loss.item():.3f}")
        
        return total_loss, gx1, gx2, gy1, gy2, mask_sum

    def fit(self, num_epochs=500,learning_rate=0.1, print_each=100):
        """
        Runs the optimization process and returns the mapping matrix.
        
        Args:
            num_epochs (int): Number of epochs to train.
            learning_rate (float): Learning rate for the optimizer.
            print_each (int): Frequency of loss printing.
        
        Returns:
            ndarray: Optimized mapping matrix (number_cells x number_spots).
        """
        torch.manual_seed(seed=self.random_state)
        optimizer = torch.optim.Adam([self.cluster_matrix], lr=learning_rate)
        
        for t in range(num_epochs):
            loss, gx1, gx2, gy1, gy2, mask_sum = self._loss_fn(verbose=(print_each is not None and t % print_each == 0))
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        
        with torch.no_grad():
            spot_matrix = softmax(self.cluster_matrix, dim=1).cpu().numpy()
        
        if self.device == 'cuda':
            torch.cuda.empty_cache()
        return spot_matrix

class Cell2VisiumSpots:
    """
    Maps each cell to its visium spot based on multi-layer spatial constraints.
    
    Attributes:
        S (torch.Tensor): Single-cell expression matrix.
        ST (AnnData): Spatial data with 3D coordinates.
        Gx (torch.Tensor): X-axis spatial expression matrix.
        Gy (torch.Tensor): Y-axis spatial expression matrix.
        Gz (torch.Tensor): Z-axis spatial expression matrix.
        lambda_gx1 (float): Weight for the first term in X-axis loss function.
        lambda_gx2 (float): Weight for the second term in X-axis loss function.
        lambda_gy1 (float): Weight for the first term in Y-axis loss function.
        lambda_gy2 (float): Weight for the second term in Y-axis loss function.
        lambda_gz1 (float): Weight for the first term in Z-axis loss function.
        lambda_gz2 (float): Weight for the second term in Z-axis loss function.
        cluster_matrix (torch.Tensor): Initial cluster matrix for optimization.
        device (str): Device to use ('cpu' or 'cuda').
        random_state (int): Random seed for initialization.
    """
    
    def __init__(
        self,
        S,
        ST,
        Gx,
        Gy,
        Gz,
        lambda_gx1=1,
        lambda_gx2=0,
        lambda_gy1=1,
        lambda_gy2=0,
        lambda_gz1=1,
        lambda_gz2=0,
        cluster_matrix=None,
        device="cpu",
        random_state=1597,
    ):
        """
        Initializes the Cell2Spots3D instance.
        
        Args:
            S (ndarray): Single-cell expression matrix.
            ST (AnnData): Spatial data with 3D coordinates.
            Gx (ndarray): X-axis spatial expression matrix.
            Gy (ndarray): Y-axis spatial expression matrix.
            Gz (ndarray): Z-axis spatial expression matrix.
            lambda_gx1 (float): Weight for the first term in X-axis loss function.
            lambda_gx2 (float): Weight for the second term in X-axis loss function.
            lambda_gy1 (float): Weight for the first term in Y-axis loss function.
            lambda_gy2 (float): Weight for the second term in Y-axis loss function.
            lambda_gz1 (float): Weight for the first term in Z-axis loss function.
            lambda_gz2 (float): Weight for the second term in Z-axis loss function.
            cluster_matrix (ndarray or None): Initial cluster matrix for optimization.
            device (str): Device to use ('cpu' or 'cuda').
            random_state (int): Random seed for initialization.
        """
        self.S = torch.tensor(S, device=device, dtype=torch.float32)
        self.ST = ST
        self.Gx = torch.tensor(Gx, device=device, dtype=torch.float32)
        self.Gy = torch.tensor(Gy, device=device, dtype=torch.float32)
        self.Gz = torch.tensor(Gz, device=device, dtype=torch.float32)
        self.lambda_gx1 = lambda_gx1
        self.lambda_gx2 = lambda_gx2
        self.lambda_gy1 = lambda_gy1
        self.lambda_gy2 = lambda_gy2
        self.lambda_gz1 = lambda_gz1
        self.lambda_gz2 = lambda_gz2
        self.random_state = random_state
        np.random.seed(seed=self.random_state)
        
        if cluster_matrix is None:
            self.cluster_matrix = np.random.normal(0, 1, (S.shape[0], ST.shape[0]))
        else:
            self.cluster_matrix = cluster_matrix
        
        self.cluster_matrix = torch.tensor(
            self.cluster_matrix, device=device, requires_grad=True, dtype=torch.float32
        )
        
        self.device = device
        self.x_length = int(ST.obs['x'].max() - ST.obs['x'].min() + 1)
        self.y_length = int(ST.obs['y'].max() - ST.obs['y'].min() + 1)
        self.z_length = int(ST.obs['z'].max() - ST.obs['z'].min() + 1)
        
        self.x_index = torch.tensor(ST.obs['x'] - ST.obs['x'].min(), device=device, dtype=torch.int32)
        self.y_index = torch.tensor(ST.obs['y'] - ST.obs['y'].min(), device=device, dtype=torch.int32)
        self.z_index = torch.tensor(ST.obs['z'] - ST.obs['z'].min(), device=device, dtype=torch.int32)

    def _generate_Xstrips(self, cluster_probs):
        """
        Generate X matrix based on regional probabilities.
        
        Args:
            cluster_probs (torch.Tensor): Cluster probabilities.
        
        Returns:
            torch.Tensor: X matrix.
        """
        mapping_x = torch.zeros((self.x_length, self.S.shape[0]), device=self.device)
        mapping_x.index_add_(0, self.x_index, cluster_probs.T)
        return mapping_x

    def _generate_Ystrips(self, cluster_probs):
        """
        Generate Y matrix based on regional probabilities.
        
        Args:
            cluster_probs (torch.Tensor): Cluster probabilities.
        
        Returns:
            torch.Tensor: Y matrix.
        """
        mapping_y = torch.zeros((self.y_length, self.S.shape[0]), device=self.device)
        mapping_y.index_add_(0, self.y_index, cluster_probs.T)
        return mapping_y
    
    def _generate_Zstrips(self, cluster_probs):
        """
        Generate Z matrix based on regional probabilities.
        
        Args:
            cluster_probs (torch.Tensor): Cluster probabilities.
        
        Returns:
            torch.Tensor: Z matrix.
        """
        mapping_z = torch.zeros((self.z_length, self.S.shape[0]), device=self.device)
        mapping_z.index_add_(0, self.z_index, cluster_probs.T)
        return mapping_z

    def _loss_fn(self, verbose=True):
        """
        Computes the loss function for the cell-to-spot mapping.
        
        Args:
            verbose (bool): If True, prints the loss.
        
        Returns:
            tuple: Total loss, and individual loss terms for X, Y, and Z axes.
        """
        cluster_probs = softmax(self.cluster_matrix, dim=1)
        
        Mx = self._generate_Xstrips(cluster_probs)
        My = self._generate_Ystrips(cluster_probs)
        Mz = self._generate_Zstrips(cluster_probs)
        
        Gx_pred = torch.matmul(Mx, self.S)
        Gy_pred = torch.matmul(My, self.S)
        Gz_pred = torch.matmul(Mz, self.S)
        
        x_gv_term = self.lambda_gx1 * cosine_similarity(Gx_pred, self.Gx, dim=0).mean()
        x_vg_term = self.lambda_gx2 * cosine_similarity(Gx_pred, self.Gx, dim=1).mean()
        y_gv_term = self.lambda_gy1 * cosine_similarity(Gy_pred, self.Gy, dim=0).mean()
        y_vg_term = self.lambda_gy2 * cosine_similarity(Gy_pred, self.Gy, dim=1).mean()
        z_gv_term = self.lambda_gz1 * cosine_similarity(Gz_pred, self.Gz, dim=0).mean()
        z_vg_term = self.lambda_gz2 * cosine_similarity(Gz_pred, self.Gz, dim=1).mean()        
        
        expression_term_x = x_gv_term + x_vg_term
        expression_term_y = y_gv_term + y_vg_term
        expression_term_z = z_gv_term + z_vg_term
        
        total_loss = -expression_term_x - expression_term_y - expression_term_z
        
        if verbose:
            print(f"Total loss: {total_loss.item():.3f}")
        
        return (
            total_loss,
            x_gv_term,
            x_vg_term,
            y_gv_term,
            y_vg_term,
            z_gv_term,
            z_vg_term
        )

    def fit(self, num_epochs=500,learning_rate=0.1, print_each=100):
        """
        Runs the optimization process and returns the mapping matrix.
        
        Args:
            num_epochs (int): Number of epochs to train.
            learning_rate (float): Learning rate for the optimizer.
            print_each (int): Frequency of loss printing.
        
        Returns:
            ndarray: Optimized mapping matrix (number_cells x number_spots).
        """
        torch.manual_seed(seed=self.random_state)
        optimizer = torch.optim.Adam([self.cluster_matrix], lr=learning_rate)
        
        for t in range(num_epochs):
            loss, *loss_terms = self._loss_fn(verbose=(print_each is not None and t % print_each == 0))
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        
        with torch.no_grad():
            spot_matrix = softmax(self.cluster_matrix, dim=1).cpu().numpy()
        
        if self.device == 'cuda':
            torch.cuda.empty_cache()
        return spot_matrix


class Cell2SpotsStrips:
    """
    Maps each cell to its spot using strips bulk RNA-seq data.
    
    Attributes:
        S (torch.Tensor): Single-cell expression matrix.
        Gx (torch.Tensor): X-axis spatial expression matrix.
        Gy (torch.Tensor): Y-axis spatial expression matrix.
        lambda_gx1 (float): Weight for the first term in X-axis loss function.
        lambda_gx2 (float): Weight for the second term in X-axis loss function.
        lambda_gy1 (float): Weight for the first term in Y-axis loss function.
        lambda_gy2 (float): Weight for the second term in Y-axis loss function.
        cluster_matrix (torch.Tensor): Initial cluster matrix for optimization.
        device (str): Device to use ('cpu' or 'cuda').
        random_state (int): Random seed for initialization.
        mask (torch.Tensor): Mask for filtering out low values.
        x_length (int): Number of X-axis spots.
        y_length (int): Number of Y-axis spots.
        x_index (torch.Tensor): Indices for X-axis spots.
        y_index (torch.Tensor): Indices for Y-axis spots.
    """
    
    def __init__(
        self,
        S,
        Gx,
        Gy,
        lambda_gx1=1,
        lambda_gx2=0,
        lambda_gy1=1,
        lambda_gy2=0,
        cluster_matrix=None,
        device="cpu",
        random_state=1597,
    ):
        """
        Initializes the Cell2SpotsStrips instance.
        
        Args:
            S (ndarray): Single-cell expression matrix.
            Gx (ndarray): X-axis spatial expression matrix.
            Gy (ndarray): Y-axis spatial expression matrix.
            lambda_gx1 (float): Weight for the first term in X-axis loss function.
            lambda_gx2 (float): Weight for the second term in X-axis loss function.
            lambda_gy1 (float): Weight for the first term in Y-axis loss function.
            lambda_gy2 (float): Weight for the second term in Y-axis loss function.
            cluster_matrix (ndarray or None): Initial cluster matrix for optimization.
            device (str): Device to use ('cpu' or 'cuda').
            random_state (int): Random seed for initialization.
        """
        self.S = torch.tensor(S, device=device, dtype=torch.float32)
        self.Gx = torch.tensor(Gx, device=device, dtype=torch.float32)
        self.Gy = torch.tensor(Gy, device=device, dtype=torch.float32)
        self.lambda_gx1 = lambda_gx1
        self.lambda_gx2 = lambda_gx2
        self.lambda_gy1 = lambda_gy1
        self.lambda_gy2 = lambda_gy2
        self.random_state = random_state
        np.random.seed(seed=self.random_state)
        
        if cluster_matrix is None:
            self.cluster_matrix = np.random.normal(0, 1, (S.shape[0], Gx.shape[1]))
        else:
            self.cluster_matrix = cluster_matrix
        
        self.cluster_matrix = torch.tensor(
            self.cluster_matrix, device=device, requires_grad=True, dtype=torch.float32
        )
        
        self.device = device
        self.x_length = Gx.shape[0]
        self.y_length = Gy.shape[0]
        
        # Assuming ST is available with x and y coordinates; otherwise, provide default indices
        self.x_index = torch.arange(self.x_length, device=device, dtype=torch.int32)
        self.y_index = torch.arange(self.y_length, device=device, dtype=torch.int32)
        
        # Mask initialization to filter out low values
        self.mask = self.cluster_matrix.clone().detach() < 1

    def _generate_Xstrips(self, cluster_probs):
        """
        Generate X matrix based on cluster probabilities.
        
        Args:
            cluster_probs (torch.Tensor): Cluster probabilities.
        
        Returns:
            torch.Tensor: X matrix.
        """
        mapping_x = torch.zeros((self.x_length, self.S.shape[0]), device=self.device)
        mapping_x.index_add_(0, self.x_index, cluster_probs.T)
        return mapping_x

    def _generate_Ystrips(self, cluster_probs):
        """
        Generate Y matrix based on cluster probabilities.
        
        Args:
            cluster_probs (torch.Tensor): Cluster probabilities.
        
        Returns:
            torch.Tensor: Y matrix.
        """
        mapping_y = torch.zeros((self.y_length, self.S.shape[0]), device=self.device)
        mapping_y.index_add_(0, self.y_index, cluster_probs.T)
        return mapping_y
    
    def _loss_fn(self, verbose=True):
        """
        Computes the loss function for cell-to-spot mapping.
        
        Args:
            verbose (bool): If True, prints the loss.
        
        Returns:
            tuple: Total loss, and individual loss terms for X and Y axes, and mask sum.
        """
        cluster_probs = softmax(self.cluster_matrix, dim=1)
        
        mask_sum = torch.masked_select(cluster_probs, self.mask).mean()
        Mx = self._generate_Xstrips(cluster_probs)
        My = self._generate_Ystrips(cluster_probs)
        
        dx_pred = torch.log(
            Mx.T.sum(axis=0) / Mx.shape[1]
        ) 
        dy_pred = torch.log(
            My.T.sum(axis=0) / My.shape[1]
        )
        
        Gx_pred = torch.matmul(Mx, self.S)
        Gy_pred = torch.matmul(My, self.S)
        
        gx1 = self.lambda_gx1 * cosine_similarity(Gx_pred, self.Gx, dim=0).mean()
        gx2 = self.lambda_gx2 * cosine_similarity(Gx_pred, self.Gx, dim=1).mean()
        gy1 = self.lambda_gy1 * cosine_similarity(Gy_pred, self.Gy, dim=0).mean()
        gy2 = self.lambda_gy2 * cosine_similarity(Gy_pred, self.Gy, dim=1).mean()
        
        cos_x = gx1 + gx2
        cos_y = gy1 + gy2
        
        main_x_loss1 = (gx1 / self.lambda_gx1).tolist()
        main_y_loss1 = (gy1 / self.lambda_gy1).tolist()
        main_x_loss2 = (gx2 / self.lambda_gx2).tolist()
        main_y_loss2 = (gy2 / self.lambda_gy2).tolist()
        
        total_loss = -cos_x - cos_y + mask_sum
        
        if verbose:
            print(f"Total loss: {total_loss.item():.3f}")
        
        return (
            total_loss,
            main_x_loss1,
            main_y_loss1,
            mask_sum
        )

    def fit(self, num_epochs=500,learning_rate=0.1, print_each=100):
        """
        Runs the optimization process and returns the mapping matrix.
        
        Args:
            num_epochs (int): Number of epochs to train.
            learning_rate (float): Learning rate for the optimizer.
            print_each (int): Frequency of loss printing.
        
        Returns:
            ndarray: Optimized mapping matrix (number_cells x number_spots).
        """
        torch.manual_seed(seed=self.random_state)
        optimizer = torch.optim.Adam([self.cluster_matrix], lr=learning_rate)
        
        for t in range(num_epochs):
            if print_each is None or t % print_each != 0:
                run_loss = self._loss_fn(verbose=False)
            else:
                run_loss = self._loss_fn(verbose=True)
            
            loss = run_loss[0]
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        
        with torch.no_grad():
            spot_matrix = softmax(self.cluster_matrix, dim=1).cpu().numpy()
        
        if self.device == 'cuda':
            torch.cuda.empty_cache()
        
        return spot_matrix


class Cell2SpotsEnhance:
    """
    Maps each cell to its spot using multi-layer custom regional information.
    
    Attributes:
        S (torch.Tensor): Single-cell expression matrix.
        ST (AnnData): Spatial transcriptomics data.
        Gx (torch.Tensor): X-axis spatial expression matrix.
        Gy (torch.Tensor): Y-axis spatial expression matrix.
        lambda_gx1 (float): Weight for the first term in X-axis loss function.
        lambda_gx2 (float): Weight for the second term in X-axis loss function.
        lambda_gy1 (float): Weight for the first term in Y-axis loss function.
        lambda_gy2 (float): Weight for the second term in Y-axis loss function.
        cluster_matrix (torch.Tensor): Initial cluster matrix for optimization.
        custom_regions (list): List of custom region labels for regional loss computation.
        lambda_r (list): List of weights for the custom regions.
        device (str): Device to use ('cpu' or 'cuda').
        random_state (int): Random seed for initialization.
        mask (torch.Tensor): Mask for filtering out low values.
        x_length (int): Number of X-axis spots.
        y_length (int): Number of Y-axis spots.
        x_index (torch.Tensor): Indices for X-axis spots.
        y_index (torch.Tensor): Indices for Y-axis spots.
    """
    
    def __init__(
        self,
        S,
        ST,
        Gx,
        Gy,
        lambda_gx1=1,
        lambda_gx2=0,
        lambda_gy1=1,
        lambda_gy2=0,
        cluster_matrix=None,
        custom_regions=None,
        lambda_r=None,
        device="cpu",
        random_state=1597,
    ):
        """
        Initializes the Cell2SpotsEnhance instance.
        
        Args:
            S (ndarray): Single-cell expression matrix.
            ST (AnnData): Spatial transcriptomics data.
            Gx (ndarray): X-axis spatial expression matrix.
            Gy (ndarray): Y-axis spatial expression matrix.
            lambda_gx1 (float): Weight for the first term in X-axis loss function.
            lambda_gx2 (float): Weight for the second term in X-axis loss function.
            lambda_gy1 (float): Weight for the first term in Y-axis loss function.
            lambda_gy2 (float): Weight for the second term in Y-axis loss function.
            cluster_matrix (ndarray or None): Initial cluster matrix for optimization.
            custom_regions (list or None): Custom regions for regional loss computation.
            lambda_r (list or None): Weights for custom regions.
            device (str): Device to use ('cpu' or 'cuda').
            random_state (int): Random seed for initialization.
        """
        self.S = torch.tensor(S, device=device, dtype=torch.float32)
        self.ST = ST
        self.Gx = torch.tensor(Gx, device=device, dtype=torch.float32)
        self.Gy = torch.tensor(Gy, device=device, dtype=torch.float32)
        self.lambda_gx1 = lambda_gx1
        self.lambda_gx2 = lambda_gx2
        self.lambda_gy1 = lambda_gy1
        self.lambda_gy2 = lambda_gy2
        self.random_state = random_state
        self.lambda_r = lambda_r if lambda_r is not None else [1] * (len(custom_regions) if custom_regions else 0)
        np.random.seed(seed=self.random_state)
        
        if cluster_matrix is None:
            self.cluster_matrix = np.random.normal(0, 1, (S.shape[0], ST.shape[0]))
            self.cluster_matrix = torch.tensor(
                self.cluster_matrix, device=device, requires_grad=True, dtype=torch.float32
            )
            self.mask = None
        else:
            self.cluster_matrix = cluster_matrix
            self.cluster_matrix = torch.tensor(
                self.cluster_matrix, device=device, requires_grad=True, dtype=torch.float32
            )
            # self.rawM = self.M.clone().detach() > 0
            self.mask = self.cluster_matrix.clone().detach() < 1
        if custom_regions is None:
            self.region = None
        else:
            self.region = custom_regions
        
        self.device = device
        self.x_length = np.int32(self.ST.obs['x'].max() - self.ST.obs['x'].min() + 1)
        self.y_length = np.int32(self.ST.obs['y'].max() - self.ST.obs['y'].min() + 1)
        self.x_index = self.ST.obs['x'] - self.ST.obs['x'].min()
        self.x_index = torch.tensor(
            self.x_index, device=device, requires_grad=False, dtype=torch.int32
        )
        self.y_index = self.ST.obs['y'] - self.ST.obs['y'].min()
        self.y_index = torch.tensor(
            self.y_index, device=device, requires_grad=False, dtype=torch.int32
        )
        
        self.region = custom_regions

    def _generate_region(self, cluster_probs, region_labels):
        """
        Generate region matrices for custom regions.
        
        Args:
            cluster_probs (torch.Tensor): Cluster probabilities.
            region_labels (list): List of custom region labels.
        
        Returns:
            tuple: List of region matrices and list of region expression matrices.
        """
        mapping_region = []
        region_exp = []
        ST_exp = torch.tensor(self.ST.X, device=self.device, dtype=torch.float32)
        
        for label in region_labels:
            st_region = torch.tensor(self.ST.obs[label].values, device=self.device, dtype=torch.float32)
            mapping_region.append(st_region.T @ cluster_probs.T)
            region_exp.append(st_region.T @ ST_exp)
        
        return mapping_region, region_exp

    def _generate_Xstrips(self, cluster_probs):
        """
        Generate X matrix based on cluster probabilities.
        
        Args:
            cluster_probs (torch.Tensor): Cluster probabilities.
        
        Returns:
            torch.Tensor: X matrix.
        """
        mapping_x = torch.zeros((self.x_length, self.S.shape[0]), device=self.device)
        mapping_x.index_add_(0, self.x_index, cluster_probs.T)
        return mapping_x

    def _generate_Ystrips(self, cluster_probs):
        """
        Generate Y matrix based on cluster probabilities.
        
        Args:
            cluster_probs (torch.Tensor): Cluster probabilities.
        
        Returns:
            torch.Tensor: Y matrix.
        """
        mapping_y = torch.zeros((self.y_length, self.S.shape[0]), device=self.device)
        mapping_y.index_add_(0, self.y_index, cluster_probs.T)
        return mapping_y

    def _loss_fn(self, verbose=True):
        """
        Computes the loss function for cell-to-spot mapping.
        
        Args:
            verbose (bool): If True, prints the loss.
        
        Returns:
            tuple: Total loss and individual loss terms for X and Y axes, and mask sum.
        """
        cluster_probs = softmax(self.cluster_matrix, dim=1)
        
        mask_sum = torch.masked_select(cluster_probs, self.mask).mean() if self.mask is not None else 0
        
        Mx = self._generate_Xstrips(cluster_probs)
        My = self._generate_Ystrips(cluster_probs)

        dx_pred = torch.log(Mx.T.sum(axis=0) / Mx.shape[1])
        dy_pred = torch.log(My.T.sum(axis=0) / My.shape[1])
        
        Gx_pred = torch.matmul(Mx, self.S)
        Gy_pred = torch.matmul(My, self.S)
        
        gx1 = self.lambda_gx1 * cosine_similarity(Gx_pred, self.Gx, dim=0).mean()
        gx2 = self.lambda_gx2 * cosine_similarity(Gx_pred, self.Gx, dim=1).mean()
        gy1 = self.lambda_gy1 * cosine_similarity(Gy_pred, self.Gy, dim=0).mean()
        gy2 = self.lambda_gy2 * cosine_similarity(Gy_pred, self.Gy, dim=1).mean()
        
        cos_x = gx1 + gx2
        cos_y = gy1 + gy2
        
        main_x_loss1 = (gx1 / self.lambda_gx1).tolist()
        main_y_loss1 = (gy1 / self.lambda_gy1).tolist()
        main_x_loss2 = (gx2 / self.lambda_gx2).tolist()
        main_y_loss2 = (gy2 / self.lambda_gy2).tolist()
        
        total_loss = -cos_x - cos_y + mask_sum
        
        if self.region:
            Mr, Er = self._generate_region(cluster_probs, self.region)
            for i in range(len(self.region)):
                Gr_pred = torch.matmul(Mr[i], self.S)
                region_loss = self.lambda_r[i] * cosine_similarity(Gr_pred, Er[i], dim=0).mean()
                total_loss -= region_loss

        if verbose:
            print(f"Total loss: {total_loss.item():.3f}")
        
        return (
            total_loss,
            main_x_loss1,
            main_y_loss1,
            mask_sum
        )

    def fit(self,  num_epochs, learning_rate=0.1, print_each=100):
        """
        Runs the optimization process and returns the mapping matrix.
        
        Args:
            num_epochs (int): Number of epochs to train.
            learning_rate (float): Learning rate for the optimizer.
            print_each (int): Frequency of loss printing.
        
        Returns:
            ndarray: Optimized mapping matrix (number_cells x number_spots).
        """
        torch.manual_seed(seed=self.random_state)
        optimizer = torch.optim.Adam([self.cluster_matrix], lr=learning_rate)
        
        for t in range(num_epochs):
            if print_each is None or t % print_each != 0:
                run_loss = self._loss_fn(verbose=False)
            else:
                run_loss = self._loss_fn(verbose=True)
                
            loss = run_loss[0]
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        
        with torch.no_grad():
            spot_matrix = softmax(self.cluster_matrix, dim=1).cpu().numpy()
            torch.cuda.empty_cache()
        
        return spot_matrix



from setuptools import setup
import setuptools
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.rst").read_text()
setup(
    name='Polyomino-sc',
    version='1.0.0',
    description='An algorithm framework employing multi-layered regional constraints to accurately assign cell locations, enhancing spatial accuracy and resilience to noise.',
    url='https://github.com/caiquanyou/Polyomino',
    author='Cai Quanyou',
    author_email='cai_quanyou@gibh.ac.cn',
    python_requires='>=3.8.0',
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        "lap>=0.4.0",
        "scikit-learn>=1.1.3",
        "numpydoc>=1.1",
        "anndata>=0.8.0",
        "scanpy>=1.9.1",
        "scipy>=1.9.3",
        "numba>=0.14.0",
        "tqdm>=4.66.5",
        "igraph>=0.11.6",
        "leidenalg>=0.10.2"
    ],
    entry_points={
        'console_scripts': [
            'polyomino=Polyomino.Polyomino:main_Polyomino',
        ]},
    classifiers=["Programming Language :: Python :: 3.8", "Operating System :: MacOS",],
    include_package_data=False
)


import numpy as np
import pandas as pd
import scanpy as sc
import anndata
from sklearn.decomposition import PCA
from scipy import sparse
from scipy.sparse import issparse
import random
from math import sin, cos, pi, sqrt
import matplotlib.pyplot as plt
import seaborn as sns
from ..extension.custom_SpaGCN import *

def rotate_coordinates(Mx, My, angle):
    """
    Rotate coordinates by a given angle.

    Args:
        Mx (np.ndarray): X-coordinates.
        My (np.ndarray): Y-coordinates.
        angle (float): Angle of rotation in radians.

    Returns:
        tuple: Rotated X and Y coordinates.
    """
    Mx_rotated = Mx * cos(angle) - My * sin(angle)
    My_rotated = Mx * sin(angle) + My * cos(angle)
    return Mx_rotated, My_rotated

def one_hot_encode_labels(labels: pd.Series):
    """
    Convert categorical labels to one-hot encoded format.

    Args:
        labels (pd.Series): Series containing categorical labels.

    Returns:
        pd.DataFrame: One-hot encoded labels.
    """
    return pd.get_dummies(labels)

def transfer_labels(mapping_matrix: np.ndarray, stdata: anndata.AnnData, labels: pd.Series):
    """
    Transfer labels from single-cell data to spatial transcriptomics data.

    Args:
        mapping_matrix (np.ndarray): Matrix mapping cells to spots.
        stdata (anndata.AnnData): Spatial transcriptomics data.
        labels (pd.Series): Cell labels.

    Returns:
        pd.DataFrame: Probability distribution of cell types for each spot.
    """
    one_hot_labels = one_hot_encode_labels(labels)
    cell_type_prob = pd.DataFrame(mapping_matrix.T @ one_hot_labels)
    cell_type_prob.index = stdata.obs.index
    return cell_type_prob

def create_spot_anndata(scdata, stdata, mapping_matrix, sclabel):
    """
    Create an AnnData object at spot resolution.

    Args:
        scdata (anndata.AnnData): Single-cell data.
        stdata (anndata.AnnData): Spatial transcriptomics data.
        mapping_matrix (np.ndarray): Matrix mapping cells to spots.
        sclabel (str): Column name of cell labels in scdata.

    Returns:
        anndata.AnnData: Spot resolution AnnData object.
    """
    spot_data = sc.AnnData(X=mapping_matrix.T @ scdata.X, var=scdata.var, obsm=stdata.obsm)
    spot_data.obsm['celltype_matrix'] = transfer_labels(mapping_matrix, stdata, scdata.obs[sclabel])
    spot_data.obsm['mapping_matrix'] = sparse.csr_matrix(mapping_matrix.T, dtype=float)
    return spot_data

def calculate_optimal_width(stdata_raw, target_spots=10000):
    """
    Calculate the optimal grid width to achieve a target number of spots.

    Args:
        stdata_raw (anndata.AnnData): Raw spatial data with single-cell resolution.
        target_spots (int): Desired number of spots.

    Returns:
        int: Optimal grid width.
    """
    spatial_coords = stdata_raw.obsm['spatial']
    stdata_raw_x = spatial_coords[:, 0] - spatial_coords[:, 0].min()
    stdata_raw_y = spatial_coords[:, 1] - spatial_coords[:, 1].min()

    x_range = stdata_raw_x.max() - stdata_raw_x.min()
    y_range = stdata_raw_y.max() - stdata_raw_y.min()

    total_area = x_range * y_range
    optimal_width = np.sqrt(total_area / target_spots)

    return int(optimal_width)

def generate_grid(stdata_raw, width=None):
    """
    Generate spatial transcriptomics data at specific grid resolution.

    Args:
        stdata_raw (anndata.AnnData): Raw spatial data with single-cell resolution.
        width (int): Width of the spots.

    Returns:
        anndata.AnnData: Spot resolution spatial transcriptomics data.
    """
    spatial_coords = stdata_raw.obsm['spatial']
    stdata_raw.obs['x'] = spatial_coords[:, 0]
    stdata_raw.obs['y'] = spatial_coords[:, 1]
    
    if width is None:
        width = calculate_optimal_width(stdata_raw)

    stdata_raw_x = stdata_raw.obs['x'] - stdata_raw.obs['x'].min()
    stdata_raw_y = stdata_raw.obs['y'] - stdata_raw.obs['y'].min()

    nx = int((stdata_raw_x.max() - stdata_raw_x.min()) // width) + 1
    ny = int((stdata_raw_y.max() - stdata_raw_y.min()) // width) + 1

    stdata_raw.obs['spot_index'] = (np.floor(stdata_raw_x / width) + np.floor(stdata_raw_y / width) * nx).astype(int)
    stdata_raw_df = stdata_raw.to_df()
    stdata_raw_df['spot_index'] = stdata_raw.obs['spot_index'].values
    stdata_raw_ep = stdata_raw_df.groupby('spot_index').sum()

    stdata = sc.AnnData(X=stdata_raw_ep.values, var=stdata_raw.var)
    stdata.obs.index = stdata_raw_ep.index.astype(str)
    stdata.obs['x'] = stdata_raw_ep.index % nx
    stdata.obs['y'] = stdata_raw_ep.index // nx
    stdata.obsm['spatial'] = stdata.obs[['x', 'y']].values

    stdata.obs['array_col'] = stdata.obsm['spatial'][:, 0].astype('int32')
    stdata.obs['array_row'] = stdata.obsm['spatial'][:, 1].astype('int32')
    stat_count = pd.value_counts(stdata_raw.obs['spot_index'])
    median_num = np.median(pd.value_counts(stdata_raw.obs['spot_index']).values)
    sns.histplot(pd.value_counts(stdata_raw.obs['spot_index']))
    plt.xlabel('Cell numbers of grid')
    plt.title('Median of cells in one grid: ' + str(median_num))
    plt.grid(None)
    return stdata

def generate_spot_matrix(stdata_raw, width):
    """
    Generate the spatial spot matrix representing cell presence in spots.

    Args:
        stdata_raw (anndata.AnnData): Raw spatial data with single-cell resolution.
        width (int): Width of the spots.

    Returns:
        np.ndarray: 3D matrix representing spatial spots.
    """
    spatial_coords = stdata_raw.obsm['spatial']
    stdata_raw.obs['x'] = spatial_coords[:, 0] - spatial_coords[:, 0].min()
    stdata_raw.obs['y'] = spatial_coords[:, 1] - spatial_coords[:, 1].min()

    nx = int((stdata_raw.obs['x'].max() - stdata_raw.obs['x'].min()) // width) + 1
    ny = int((stdata_raw.obs['y'].max() - stdata_raw.obs['y'].min()) // width) + 1

    spot_matrix = np.zeros((len(stdata_raw), nx * ny), dtype=int)

    for i, cell in enumerate(stdata_raw):
        x, y = cell.obs['x'].values, cell.obs['y'].values
        spot_idx = int(np.floor(x / width) + np.floor(y / width) * nx)
        spot_matrix[i, spot_idx] = 1

    spot_matrix_reshaped = spot_matrix.reshape((len(stdata_raw), ny, nx))
    return spot_matrix_reshaped[:, ::-1, :]  # Reverse the y-axis to match the typical plot orientation

def generate_empty_grid(stdata_raw, stdata):
    """
    Generate an empty grid matrix based on spatial data.

    Args:
        stdata_raw (anndata.AnnData): Raw spatial data with single-cell resolution.
        stdata (anndata.AnnData): Spot resolution spatial data.

    Returns:
        np.ndarray: Empty grid matrix.
    """
    grid_shape = (int(stdata.obs['y'].max()) + 1, int(stdata.obs['x'].max()) + 1)
    empty_grid = np.zeros(grid_shape, dtype=int)

    for idx in stdata_raw.obs['spot_index'].unique():
        empty_grid.flat[int(idx)] = 1

    return empty_grid

def generate_Xstrips(stdata):
    """this function is to generate X strips, which is used for calculating the correlation matrix
    Args:
        stdata (anndata.AnnData): Visium data
    Returns:
        the function will return the X strips and update the 'x' column in stdata.obs
    """
    slides_x =  []
    x_min = stdata.obs['array_col'].min()
    x_max = stdata.obs['array_col'].max()
    lenth = x_max-x_min+1
    j = 0
    for i in range(np.int32(lenth)):
        idx = abs(stdata.obs['array_col'] - x_min - i) < 0.1
        if sum(idx)==0:
            continue
        slides_x.append(np.asarray(stdata.X[idx,:].sum(axis =0)).reshape(-1)) 
        stdata.obs.loc[idx,'x'] = j
        j+=1
    return np.array(slides_x)

def generate_Ystrips(stdata):
    """this function is to generate Y strips, which is used for calculating the correlation matrix
    Args:
        stdata (anndata.AnnData): Visium data
    Returns:
        the function will return the Y strips and update the 'y' column in stdata.obs
    """
    slides_y =  []
    y_min = stdata.obs['array_row'].min()
    y_max = stdata.obs['array_row'].max()
    lenth = y_max-y_min+1
    j = 0
    for i in range(np.int32(lenth)):
        idx = abs(stdata.obs['array_row'] - y_min - i) < 0.1
        if sum(idx)==0:
            continue
        slides_y.append(np.asarray(stdata.X[idx,:].sum(axis =0)).reshape(-1)) 
        stdata.obs.loc[idx,'y'] = j
        j+=1
    return np.array(slides_y)

def generate_Zstrips(stdata):
    """Generate X strips
    Args:
        stdata (anndata): Visium data
    Returns:
        BYn: _description_
    """
    slides_z =  []
    # Mx_p,My_p = rotation_exp(stdata.obs['array_col'],stdata.obs['array_row'],angle=pi/4)
    Mx_p,My_p = rotation_exp(stdata.obs['array_col'],stdata.obs['array_row'],angle=pi/4)
    lenth = np.round((Mx_p.max()-Mx_p.min())/sqrt(2))+1
    start_value = Mx_p.min()
    j = 0 
    for i in range(np.int32(lenth)):
        idx = abs(Mx_p - start_value - i * sqrt(2)) < 0.1
        if sum(idx)==0:
            continue
        slides_z.append(np.asarray(stdata.X[idx,:].sum(axis =0)).reshape(-1)) 
        stdata.obs.loc[idx,'z'] = j
        j += 1
    return np.array(slides_z)

def generate_cluster_exp(stdata, label):
    """
    Generate a cluster expression matrix.

    Args:
        stdata (anndata.AnnData): Spatial matrix containing the data.
        label (str): The column name of the clusters label stored in the spatial data's `.obs`.

    Returns:
        np.ndarray: A matrix where each row corresponds to the sum of expression values for each cluster.
    """
    clusters = []
    for cluster_label in stdata.obs[label].unique():
        cluster_data = stdata[stdata.obs[label] == cluster_label]
        clusters.append(np.asarray(cluster_data.X.sum(axis=0)).reshape(-1))
    return np.array(clusters)


def preprocess(scdata, stdata, genes=None):
    """
    Preprocess the single-cell and spatial data by filtering out genes not expressed in any cells and identifying overlapping genes.

    Args:
        scdata (anndata.AnnData): Single-cell RNA-seq data.
        stdata (anndata.AnnData): Spatial transcriptomics data.
        genes (list, optional): List of genes to consider. If None, all genes are considered.

    Returns:
        None: The function updates `scdata.uns["overlap_genes"]` with the overlapping genes.
    """
    sc.pp.filter_genes(scdata, min_cells=1)
    if genes is None:
        genes = scdata.var.index
    overlap_genes = list(set(genes) & set(scdata.var.index) & set(stdata.var.index))
    scdata.uns["overlap"] = overlap_genes


def random_cluster(stdata, label, embed=None, n_pcs=50, pc_frac=0.5, samples_time=1, shape="square", target_num=10):
    """
    Generate random cluster labels using PCA and SpaGCN clustering.

    Args:
        stdata (anndata.AnnData): Spatial matrix containing the data.
        label (str): Column name to store the generated cluster labels in `stdata.obs`.
        embed (np.ndarray, optional): Precomputed embeddings. If None, PCA will be performed.
        n_pcs (int, optional): Number of principal components to use for clustering. Default is 50.
        pc_frac (float, optional): Fraction of PCs to randomly select. Default is 0.5.
        samples_time (int, optional): Number of times to sample. Default is 1.
        shape (str, optional): Shape of the spot. Default is "square". Other options are "hex" for Visium data.
        target_num (int, optional): Target number of clusters. Default is 10.

    Returns:
        None: The function updates `stdata.obs` with the generated cluster labels.
    """
    if embed is None:
        pca = PCA(n_components=n_pcs)
        if issparse(stdata.X):
            pca.fit(stdata.X.A)
            embed = pca.transform(stdata.X.A)
        else:
            pca.fit(stdata.X)
            embed = pca.transform(stdata.X)
    
    selected_pcs_indices = sorted(random.sample(range(n_pcs), int(n_pcs * pc_frac)))
    selected_pcs = embed[:, selected_pcs_indices]
    SpaGCN_cluster(selected_pcs, stdata, label, shape=shape, target_num=target_num)


def SpaGCN_cluster(selected_pcs, stdata, label, shape="square", target_num=10):
    """
    Perform clustering using SpaGCN on the selected principal components.

    Args:
        selected_pcs (np.ndarray): Selected principal components for clustering.
        stdata (anndata.AnnData): Spatial matrix containing the data.
        label (str): Column name to store the generated cluster labels in `stdata.obs`.
        shape (str, optional): Shape of the spot. Default is "square". Other options are "hex" for Visium data.
        target_num (int, optional): Target number of clusters. Default is 10.

    Returns:
        None: The function updates `stdata.obs` with the refined cluster labels.
    """
    prefilter_specialgenes(stdata)

    x_pixel = stdata.obs["x"].tolist()
    y_pixel = stdata.obs["y"].tolist()
    adj_no_img = calculate_adj_matrix(x=x_pixel, y=y_pixel, histology=False)
    
    l = search_l(p=0.5, adj=adj_no_img, start=0.01, end=1000, tol=0.01, max_run=100)
    
    r_seed = t_seed = n_seed = 100
    res = search_res(
        stdata, adj_no_img, selected_pcs, l,
        target_num=target_num, start=0.7, step=0.1, tol=5e-3, lr=0.05, max_epochs=20,
        r_seed=r_seed, t_seed=t_seed, n_seed=n_seed
    )
    
    clf = SpaGCN()
    clf.set_l(l)
    random.seed(r_seed)
    torch.manual_seed(t_seed)
    np.random.seed(n_seed)
    
    clf.train(
        stdata, adj_no_img, selected_pcs,
        init_spa=True, init="leiden", res=res, tol=5e-3, lr=0.05, max_epochs=200
    )
    y_pred, prob = clf.predict()
    stdata.obs[label]= y_pred
    stdata.obs[label]=stdata.obs[label].astype('category')
    refined_pred=refine(sample_id=stdata.obs.index.tolist(), pred=stdata.obs[label].tolist(), dis=adj_no_img, shape=shape)
    stdata.obs[label]=refined_pred
    stdata.obs[label]=stdata.obs[label].astype('category')
    
def sc_accu(recon_data, raw_st, recon_label, raw_label, nn=5, thres=1):
    """
    Calculate the accuracy of the reconstructed cell types by comparing them to the raw spatial data.

    Args:
        recon_data (anndata.AnnData): Reconstructed spatial data.
        raw_st (anndata.AnnData): Raw spatial data.
        recon_label (str): Column name for the reconstructed cell type labels in `recon_data.obs`.
        raw_label (str): Column name for the true cell type labels in `raw_st.obs`.
        nn (int, optional): Number of nearest neighbors to consider. Default is 5.
        thres (int, optional): Threshold for the number of matches required to consider a cell type correct. Default is 1.

    Returns:
        pd.DataFrame: A DataFrame containing the neighbor analysis and match counts.
    """
    from sklearn.neighbors import KDTree

    # Create KDTree for nearest neighbor search
    kdt = KDTree(recon_data.obsm['spatial'], leaf_size=30, metric='euclidean')
    nn_result = kdt.query(recon_data.obsm['spatial'], k=nn, return_distance=False)

    # Create a mapping from the raw spatial data indices to cell type labels
    mapping_dict = dict(zip(range(len(raw_st)), raw_st.obs[raw_label]))

    # Create a DataFrame with neighbor indices and map them to true cell types
    column_names = [f"neighbor_{i}" for i in range(nn)]
    results = pd.DataFrame(nn_result, columns=column_names).applymap(mapping_dict.get)

    # Add reconstructed and true cell type labels to the DataFrame
    results['Recon_ct'] = recon_data.obs[recon_label].values
    results['True_ct'] = raw_st.obs[raw_label].values

    # Calculate the number of matches with the reconstructed cell type
    results['Matches'] = results[column_names].eq(results['Recon_ct'], axis=0).sum(axis=1)

    # Determine if the number of matches meets or exceeds the threshold
    results['Values_Equal'] = results['Matches'] >= thres

    return results


def fast_pearson(v1, v2):
    """
    Compute Pearson correlation coefficients quickly between two sets of vectors.

    Args:
        v1 (np.ndarray): First set of vectors, shape (n_samples, n_features).
        v2 (np.ndarray): Second set of vectors, shape (n_samples, n_features).

    Returns:
        np.ndarray: Pearson correlation coefficients matrix.
    """
    n = v1.shape[0]
    sums = np.multiply.outer(v2.sum(axis=0), v1.sum(axis=0))
    stds = np.multiply.outer(v2.std(axis=0), v1.std(axis=0))
    correlation = (v2.T.dot(v1) - sums / n) / stds / n

    return correlation


def fast_spearman(v1, v2):
    """
    Compute Spearman correlation coefficients quickly between two sets of vectors.

    Args:
        v1 (np.ndarray): First set of vectors, shape (n_samples, n_features).
        v2 (np.ndarray): Second set of vectors, shape (n_samples, n_features).

    Returns:
        np.ndarray: Spearman correlation coefficients matrix.
    """
    v1 = pd.DataFrame(v1).rank().values
    v2 = pd.DataFrame(v2).rank().values
    n = v1.shape[0]
    sums = np.multiply.outer(v2.sum(axis=0), v1.sum(axis=0))
    stds = np.multiply.outer(v2.std(axis=0), v1.std(axis=0))
    correlation = (v2.T.dot(v1) - sums / n) / stds / n

    return correlation



__version__ = '1.0.0'
import warnings
warnings.filterwarnings('ignore')

from .utilities.util import *
from .model.ot_model import *
from .model.model import *
from .extension.custom_SpaGCN import *
from . import plotting as pl

from .Polyomino import *
from .Polyomino_argument import *
